import torch
import numpy as np

from .galaxy_model_object import Galaxy_Model
from .psf_model_object import PSF_Model
from .warp_model import Warp_Galaxy
from .ray_model import Ray_Galaxy
from .wedge_model import Wedge_Galaxy
from .superellipse_model import SuperEllipse_Galaxy, SuperEllipse_Warp
from .foureirellipse_model import FourierEllipse_Galaxy, FourierEllipse_Warp
from ._shared_methods import (
    parametric_initialize,
    parametric_segment_initialize,
    select_target,
)
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.parametric_profiles import nuker_np

__all__ = [
    "Nuker_Galaxy",
    "Nuker_PSF",
    "Nuker_SuperEllipse",
    "Nuker_SuperEllipse_Warp",
    "Nuker_FourierEllipse",
    "Nuker_FourierEllipse_Warp",
    "Nuker_Warp",
    "Nuker_Ray",
]


def _x0_func(model_params, R, F):
    return R[4], F[4], 1.0, 2.0, 0.5


def _wrap_nuker(R, rb, ib, a, b, g):
    return nuker_np(R, rb, 10 ** (ib), a, b, g)


class Nuker_Galaxy(Galaxy_Model):
    """basic galaxy model with a Nuker profile for the radial light
    profile. The functional form of the Nuker profile is defined as:

    I(R) = Ib * 2^((beta-gamma)/alpha) * (R / Rb)^(-gamma) * (1 + (R/Rb)^alpha)^((gamma - beta)/alpha)

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ib is the flux density at
    the scale radius Rb, Rb is the scale length for the profile, beta
    is the outer power law slope, gamma is the iner power law slope,
    and alpha is the sharpness of the transition.

    Parameters:
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """

    model_type = f"nuker {Galaxy_Model.model_type}"
    parameter_specs = {
        "Rb": {"units": "arcsec", "limits": (0, None)},
        "Ib": {"units": "log10(flux/arcsec^2)"},
        "alpha": {"units": "none", "limits": (0, None)},
        "beta": {"units": "none", "limits": (0, None)},
        "gamma": {"units": "none"},
    }
    _parameter_order = Galaxy_Model._parameter_order + (
        "Rb",
        "Ib",
        "alpha",
        "beta",
        "gamma",
    )
    useable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(
            self,
            parameters,
            target,
            _wrap_nuker,
            ("Rb", "Ib", "alpha", "beta", "gamma"),
            _x0_func,
        )

    from ._shared_methods import nuker_radial_model as radial_model


class Nuker_PSF(PSF_Model):
    """basic point source model with a Nuker profile for the radial light
    profile. The functional form of the Nuker profile is defined as:

    I(R) = Ib * 2^((beta-gamma)/alpha) * (R / Rb)^(-gamma) * (1 + (R/Rb)^alpha)^((gamma - beta)/alpha)

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ib is the flux density at
    the scale radius Rb, Rb is the scale length for the profile, beta
    is the outer power law slope, gamma is the iner power law slope,
    and alpha is the sharpness of the transition.

    Parameters:
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """

    model_type = f"nuker {PSF_Model.model_type}"
    parameter_specs = {
        "Rb": {"units": "arcsec", "limits": (0, None)},
        "Ib": {"units": "log10(flux/arcsec^2)", "value": 0., "locked": True},
        "alpha": {"units": "none", "limits": (0, None)},
        "beta": {"units": "none", "limits": (0, None)},
        "gamma": {"units": "none"},
    }
    _parameter_order = PSF_Model._parameter_order + (
        "Rb",
        "Ib",
        "alpha",
        "beta",
        "gamma",
    )
    useable = True
    model_integrated = False
    
    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(
            self,
            parameters,
            target,
            _wrap_nuker,
            ("Rb", "Ib", "alpha", "beta", "gamma"),
            _x0_func,
        )

    from ._shared_methods import nuker_radial_model as radial_model
    from ._shared_methods import radial_evaluate_model as evaluate_model


class Nuker_SuperEllipse(SuperEllipse_Galaxy):
    """super ellipse galaxy model with a Nuker profile for the radial
    light profile. The functional form of the Nuker profile is defined as:

    I(R) = Ib * 2^((beta-gamma)/alpha) * (R / Rb)^(-gamma) * (1 + (R/Rb)^alpha)^((gamma - beta)/alpha)

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ib is the flux density at
    the scale radius Rb, Rb is the scale length for the profile, beta
    is the outer power law slope, gamma is the iner power law slope,
    and alpha is the sharpness of the transition.

    Parameters:
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """

    model_type = f"nuker {SuperEllipse_Galaxy.model_type}"
    parameter_specs = {
        "Rb": {"units": "arcsec", "limits": (0, None)},
        "Ib": {"units": "log10(flux/arcsec^2)"},
        "alpha": {"units": "none", "limits": (0, None)},
        "beta": {"units": "none", "limits": (0, None)},
        "gamma": {"units": "none"},
    }
    _parameter_order = SuperEllipse_Galaxy._parameter_order + (
        "Rb",
        "Ib",
        "alpha",
        "beta",
        "gamma",
    )
    useable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(
            self,
            parameters,
            target,
            _wrap_nuker,
            ("Rb", "Ib", "alpha", "beta", "gamma"),
            _x0_func,
        )

    from ._shared_methods import nuker_radial_model as radial_model


class Nuker_SuperEllipse_Warp(SuperEllipse_Warp):
    """super ellipse warp galaxy model with a Nuker profile for the
    radial light profile. The functional form of the Nuker profile is
    defined as:

    I(R) = Ib * 2^((beta-gamma)/alpha) * (R / Rb)^(-gamma) * (1 + (R/Rb)^alpha)^((gamma - beta)/alpha)

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ib is the flux density at
    the scale radius Rb, Rb is the scale length for the profile, beta
    is the outer power law slope, gamma is the iner power law slope,
    and alpha is the sharpness of the transition.

    Parameters:
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope


    """

    model_type = f"nuker {SuperEllipse_Warp.model_type}"
    parameter_specs = {
        "Rb": {"units": "arcsec", "limits": (0, None)},
        "Ib": {"units": "log10(flux/arcsec^2)"},
        "alpha": {"units": "none", "limits": (0, None)},
        "beta": {"units": "none", "limits": (0, None)},
        "gamma": {"units": "none"},
    }
    _parameter_order = SuperEllipse_Warp._parameter_order + (
        "Rb",
        "Ib",
        "alpha",
        "beta",
        "gamma",
    )
    useable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(
            self,
            parameters,
            target,
            _wrap_nuker,
            ("Rb", "Ib", "alpha", "beta", "gamma"),
            _x0_func,
        )

    from ._shared_methods import nuker_radial_model as radial_model


class Nuker_FourierEllipse(FourierEllipse_Galaxy):
    """fourier mode perturbations to ellipse galaxy model with a Nuker
    profile for the radial light profile. The functional form of the
    Nuker profile is defined as:

    I(R) = Ib * 2^((beta-gamma)/alpha) * (R / Rb)^(-gamma) * (1 + (R/Rb)^alpha)^((gamma - beta)/alpha)

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ib is the flux density at
    the scale radius Rb, Rb is the scale length for the profile, beta
    is the outer power law slope, gamma is the iner power law slope,
    and alpha is the sharpness of the transition.

    Parameters:
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """

    model_type = f"nuker {FourierEllipse_Galaxy.model_type}"
    parameter_specs = {
        "Rb": {"units": "arcsec", "limits": (0, None)},
        "Ib": {"units": "log10(flux/arcsec^2)"},
        "alpha": {"units": "none", "limits": (0, None)},
        "beta": {"units": "none", "limits": (0, None)},
        "gamma": {"units": "none"},
    }
    _parameter_order = FourierEllipse_Galaxy._parameter_order + (
        "Rb",
        "Ib",
        "alpha",
        "beta",
        "gamma",
    )
    useable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(
            self,
            parameters,
            target,
            _wrap_nuker,
            ("Rb", "Ib", "alpha", "beta", "gamma"),
            _x0_func,
        )

    from ._shared_methods import nuker_radial_model as radial_model


class Nuker_FourierEllipse_Warp(FourierEllipse_Warp):
    """fourier mode perturbations to ellipse galaxy model with a Nuker
    profile for the radial light profile. The functional form of the
    Nuker profile is defined as:

    I(R) = Ib * 2^((beta-gamma)/alpha) * (R / Rb)^(-gamma) * (1 + (R/Rb)^alpha)^((gamma - beta)/alpha)

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ib is the flux density at
    the scale radius Rb, Rb is the scale length for the profile, beta
    is the outer power law slope, gamma is the iner power law slope,
    and alpha is the sharpness of the transition.

    Parameters:
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """

    model_type = f"nuker {FourierEllipse_Warp.model_type}"
    parameter_specs = {
        "Rb": {"units": "arcsec", "limits": (0, None)},
        "Ib": {"units": "log10(flux/arcsec^2)"},
        "alpha": {"units": "none", "limits": (0, None)},
        "beta": {"units": "none", "limits": (0, None)},
        "gamma": {"units": "none"},
    }
    _parameter_order = FourierEllipse_Warp._parameter_order + (
        "Rb",
        "Ib",
        "alpha",
        "beta",
        "gamma",
    )
    useable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(
            self,
            parameters,
            target,
            _wrap_nuker,
            ("Rb", "Ib", "alpha", "beta", "gamma"),
            _x0_func,
        )

    from ._shared_methods import nuker_radial_model as radial_model


class Nuker_Warp(Warp_Galaxy):
    """warped coordinate galaxy model with a Nuker profile for the radial
    light model. The functional form of the Nuker profile is defined
    as:

    I(R) = Ib * 2^((beta-gamma)/alpha) * (R / Rb)^(-gamma) * (1 + (R/Rb)^alpha)^((gamma - beta)/alpha)

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ib is the flux density at
    the scale radius Rb, Rb is the scale length for the profile, beta
    is the outer power law slope, gamma is the iner power law slope,
    and alpha is the sharpness of the transition.

    Parameters:
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """

    model_type = f"nuker {Warp_Galaxy.model_type}"
    parameter_specs = {
        "Rb": {"units": "arcsec", "limits": (0, None)},
        "Ib": {"units": "log10(flux/arcsec^2)"},
        "alpha": {"units": "none", "limits": (0, None)},
        "beta": {"units": "none", "limits": (0, None)},
        "gamma": {"units": "none"},
    }
    _parameter_order = Warp_Galaxy._parameter_order + (
        "Rb",
        "Ib",
        "alpha",
        "beta",
        "gamma",
    )
    useable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(
            self,
            parameters,
            target,
            _wrap_nuker,
            ("Rb", "Ib", "alpha", "beta", "gamma"),
            _x0_func,
        )

    from ._shared_methods import nuker_radial_model as radial_model


class Nuker_Ray(Ray_Galaxy):
    """ray galaxy model with a nuker profile for the radial light
    model. The functional form of the Sersic profile is defined as:

    I(R) = Ib * 2^((beta-gamma)/alpha) * (R / Rb)^(-gamma) * (1 + (R/Rb)^alpha)^((gamma - beta)/alpha)

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ib is the flux density at
    the scale radius Rb, Rb is the scale length for the profile, beta
    is the outer power law slope, gamma is the iner power law slope,
    and alpha is the sharpness of the transition.

    Parameters:
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """

    model_type = f"nuker {Ray_Galaxy.model_type}"
    parameter_specs = {
        "Rb": {"units": "arcsec", "limits": (0, None)},
        "Ib": {"units": "log10(flux/arcsec^2)"},
        "alpha": {"units": "none", "limits": (0, None)},
        "beta": {"units": "none", "limits": (0, None)},
        "gamma": {"units": "none"},
    }
    _parameter_order = Ray_Galaxy._parameter_order + (
        "Rb",
        "Ib",
        "alpha",
        "beta",
        "gamma",
    )
    useable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_segment_initialize(
            model=self,
            parameters=parameters,
            target=target,
            prof_func=_wrap_nuker,
            params=("Rb", "Ib", "alpha", "beta", "gamma"),
            x0_func=_x0_func,
            segments=self.rays,
        )

    from ._shared_methods import nuker_iradial_model as iradial_model


class Nuker_Wedge(Wedge_Galaxy):
    """wedge galaxy model with a nuker profile for the radial light
    model. The functional form of the Sersic profile is defined as:

    I(R) = Ib * 2^((beta-gamma)/alpha) * (R / Rb)^(-gamma) * (1 + (R/Rb)^alpha)^((gamma - beta)/alpha)

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ib is the flux density at
    the scale radius Rb, Rb is the scale length for the profile, beta
    is the outer power law slope, gamma is the iner power law slope,
    and alpha is the sharpness of the transition.

    Parameters:
        Ib: brightness at the scale length, represented as the log of the brightness divided by pixel scale squared.
        Rb: scale length radius
        alpha: sharpness of transition between power law slopes
        beta: outer power law slope
        gamma: inner power law slope

    """

    model_type = f"nuker {Wedge_Galaxy.model_type}"
    parameter_specs = {
        "Rb": {"units": "arcsec", "limits": (0, None)},
        "Ib": {"units": "log10(flux/arcsec^2)"},
        "alpha": {"units": "none", "limits": (0, None)},
        "beta": {"units": "none", "limits": (0, None)},
        "gamma": {"units": "none"},
    }
    _parameter_order = Wedge_Galaxy._parameter_order + (
        "Rb",
        "Ib",
        "alpha",
        "beta",
        "gamma",
    )
    useable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_segment_initialize(
            model=self,
            parameters=parameters,
            target=target,
            prof_func=_wrap_nuker,
            params=("Rb", "Ib", "alpha", "beta", "gamma"),
            x0_func=_x0_func,
            segments=self.wedges,
        )

    from ._shared_methods import nuker_iradial_model as iradial_model



import unittest
import astrophot as ap
import torch
import numpy as np
from utils import make_basic_sersic, make_basic_gaussian_psf
#torch.autograd.set_detect_anomaly(True)
######################################################################
# PSF Model Objects
######################################################################

class TestAllPSFModelBasics(unittest.TestCase):
    def test_all_psfmodel_sample(self):

        target = make_basic_gaussian_psf()
        for model_type in ap.models.PSF_Model.List_Model_Names(useable=True):
            print(model_type)
            MODEL = ap.models.AstroPhot_Model(
                name="test model",
                model_type=model_type,
                target=target,
            )
            MODEL.initialize()
            for P in MODEL.parameter_order:
                self.assertIsNotNone(
                    MODEL[P].value,
                    f"Model type {model_type} parameter {P} should not be None after initialization",
                )
            print(MODEL.parameters)
            img = MODEL()
            self.assertTrue(
                torch.all(torch.isfinite(img.data)),
                "Model should evaluate a real number for the full image",
            )
            self.assertIsInstance(str(MODEL), str, "String representation should return string")
            self.assertIsInstance(repr(MODEL), str, "Repr should return string")

class TestEigenPSF(unittest.TestCase):
    def test_init(self):
        target = make_basic_gaussian_psf(N = 51, rand = 666)
        dat = target.data.detach()
        dat[dat < 0] = 0
        target = ap.image.PSF_Image(data = dat, pixelscale = target.pixelscale)
        basis = np.stack(list(make_basic_gaussian_psf(N = 51, sigma = s, rand = int(4923*s)).data for s in np.linspace(8, 1, 5)))
        # basis = np.random.rand(10,51,51)
        EM = ap.models.AstroPhot_Model(
            model_type = "eigen psf model",
            eigen_basis = basis,
            eigen_pixelscale = 1,
            target = target,
        )

        EM.initialize()

        res = ap.fit.LM(EM, verbose = 1).fit()

        self.assertEqual(res.message, "success")

class TestPixelPSF(unittest.TestCase):
    def test_init(self):
        target = make_basic_gaussian_psf(N = 11)
        target.data[target.data < 0] = 0
        target = ap.image.PSF_Image(data = target.data / torch.sum(target.data), pixelscale = target.pixelscale)
        
        PM = ap.models.AstroPhot_Model(
            model_type = "pixelated psf model",
            target = target,
        )

        PM.initialize()
        
        self.assertTrue(torch.allclose(PM().data, target.data))



import unittest
from astrophot import image
import astrophot as ap
import torch

from utils import get_astropy_wcs

######################################################################
# Image_Header Objects
######################################################################

class TestImageHeader(unittest.TestCase):
    def test_image_header_creation(self):

        # Minimial input
        H = ap.image.Image_Header(
            data_shape = (20,20),
            zeropoint = 22.5,
            pixelscale = 0.2,
        )

        self.assertTrue(torch.all(H.origin == 0), "Origin should be assumed zero if not given")

        # Center
        H = ap.image.Image_Header(
            data_shape = (20,20),
            pixelscale = 0.2,
            center = (10,10),
        )

        self.assertTrue(torch.all(H.origin == 8), "Center provided, origin should be adjusted accordingly")

        # Origin
        H = ap.image.Image_Header(
            data_shape = (20,20),
            pixelscale = 0.2,
            origin = (10,10),
        )

        self.assertTrue(torch.all(H.origin == 10), "Origin provided, origin should be as given")

        # Center radec
        H = ap.image.Image_Header(
            data_shape = (20,20),
            pixelscale = 0.2,
            center_radec = (10,10),
        )

        self.assertTrue(torch.allclose(H.plane_to_world(H.center), torch.ones_like(H.center) * 10), "Center_radec provided, center should be as given in world coordinates")

        # Origin radec
        H = ap.image.Image_Header(
            data_shape = (20,20),
            pixelscale = 0.2,
            origin_radec = (10,10),
        )

        self.assertTrue(torch.allclose(H.plane_to_world(H.origin), torch.ones_like(H.center) * 10), "Origin_radec provided, origin should be as given in world coordinates")

        # Astropy WCS
        wcs = get_astropy_wcs()
        H = ap.image.Image_Header(
            data_shape = (180,180),
            wcs = wcs,
        )

        sky_coord = wcs.pixel_to_world(*wcs.wcs.crpix)
        wcs_world = torch.tensor((sky_coord.ra.deg, sky_coord.dec.deg))
        self.assertTrue(torch.allclose(torch.tensor(wcs.wcs.crpix, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device), H.world_to_pixel(wcs_world)), "Astropy WCS initialization should map crval crpix coordinates")
        
        
    def test_image_header_wcs_roundtrip(self):

        wcs = get_astropy_wcs()
        # Minimial input
        H = ap.image.Image_Header(
            data_shape = (20,20),
            zeropoint = 22.5,
            wcs = wcs,
        )

        self.assertTrue(torch.allclose(H.world_to_plane(H.plane_to_world(torch.zeros_like(H.window.reference_radec))), torch.zeros_like(H.window.reference_radec)), "WCS world/plane roundtrip should return input value")
        self.assertTrue(torch.allclose(H.pixel_to_plane(H.plane_to_pixel(torch.zeros_like(H.window.reference_radec))), torch.zeros_like(H.window.reference_radec)), "WCS pixel/plane roundtrip should return input value")
        self.assertTrue(torch.allclose(H.world_to_pixel(H.pixel_to_world(torch.zeros_like(H.window.reference_radec))), torch.zeros_like(H.window.reference_radec), atol = 1e-6), "WCS world/pixel roundtrip should return input value")

        self.assertTrue(torch.allclose(H.pixel_to_plane_delta(H.plane_to_pixel_delta(torch.ones_like(H.window.reference_radec))), torch.ones_like(H.window.reference_radec)), "WCS pixel/plane delta roundtrip should return input value")

        
        
    def test_iamge_header_repr(self):

        wcs = get_astropy_wcs()
        # Minimial input
        H = ap.image.Image_Header(
            data_shape = (20,20),
            zeropoint = 22.5,
            wcs = wcs,
        )

        S = str(H)
        R = repr(H)
        



import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import iqr
from scipy.fftpack import fft

from ..isophote.extract import _iso_extract


def isophotes(
    image, center, threshold=None, pa=None, q=None, R=None, n_isophotes=3, more=False
):
    """Method for quickly extracting a small number of elliptical
    isophotes for the sake of initializing other models.

    """

    if pa is None:
        pa = 0.0

    if q is None:
        q = 1.0

    if R is None:
        # Determine basic threshold if none given
        if threshold is None:
            threshold = np.nanmedian(image) + 3 * iqr(image[np.isfinite(image)], rng=(16, 84)) / 2
            
        # Sample growing isophotes until threshold is reached
        ellipse_radii = [1.0]
        while ellipse_radii[-1] < (max(image.shape) / 2):
            ellipse_radii.append(ellipse_radii[-1] * (1 + 0.2))
            isovals = _iso_extract(
                image,
                ellipse_radii[-1],
                {
                    "q": q if isinstance(q, float) else np.max(q),
                    "pa": pa if isinstance(pa, float) else np.min(pa),
                },
                {"x": center[0], "y": center[1]},
                more=False,
                sigmaclip=True,
                sclip_nsigma=3,
            )
            if len(isovals) < 3:
                continue
            # Stop when at 3 time background noise
            if (np.quantile(isovals, 0.8) < threshold) and len(ellipse_radii) > 4:
                break
        R = ellipse_radii[-1]
        
    # Determine which radii to sample based on input R, pa, and q
    if isinstance(pa, float) and isinstance(q, float) and isinstance(R, float):
        if n_isophotes == 1:
            isophote_radii = [R]
        else:
            isophote_radii = np.linspace(0, R, n_isophotes)
    elif hasattr(R, "__len__"):
        isophote_radii = R
    elif hasattr(pa, "__len__"):
        isophote_radii = np.ones(len(pa)) * R
    elif hasattr(q, "__len__"):
        isophote_radii = np.ones(len(q)) * R
        
    # Sample the requested isophotes and record desired info
    iso_info = []
    for i, r in enumerate(isophote_radii):
        iso_info.append({"R": r})
        isovals = _iso_extract(
            image,
            r,
            {
                "q": q if isinstance(q, float) else q[i],
                "pa": pa if isinstance(pa, float) else pa[i],
            },
            {"x": center[0], "y": center[1]},
            more=more,
            sigmaclip=True,
            sclip_nsigma=3,
            interp_mask=True,
        )
        if more:
            angles = isovals[1]
            isovals = isovals[0]
        if len(isovals) < 3:
            iso_info[-1] = None
            continue
        coefs = fft(isovals)
        iso_info[-1]["phase1"] = np.angle(coefs[1])
        iso_info[-1]["phase2"] = np.angle(coefs[2])
        iso_info[-1]["flux"] = np.median(isovals)
        iso_info[-1]["noise"] = iqr(isovals, rng=(16, 84)) / 2
        iso_info[-1]["amplitude1"] = np.abs(coefs[1]) / (
            len(isovals) * (max(0, iso_info[-1]["flux"]) + iso_info[-1]["noise"])
        )
        iso_info[-1]["amplitude2"] = np.abs(coefs[2]) / (
            len(isovals) * (max(0, iso_info[-1]["flux"]) + iso_info[-1]["noise"])
        )
        iso_info[-1]["N"] = len(isovals)
        if more:
            iso_info[-1]["isovals"] = isovals
            iso_info[-1]["angles"] = angles

    # recover lost isophotes just to keep code moving
    for i in reversed(range(len(iso_info))):
        if iso_info[i] is not None:
            good_index = i
            break
    else:
        raise ValueError(
            "Unable to recover any isophotes, try on a better band or manually provide values"
        )
    for i in range(len(iso_info)):
        if iso_info[i] is None:
            iso_info[i] = iso_info[good_index]
            iso_info[i]["R"] = isophote_radii[i]
    return iso_info



import sys
import argparse
import requests
import torch
from .parse_config import galfit_config, basic_config
from . import models, image, plots, utils, fit, param, AP_config

try:
    from ._version import version as VERSION  # noqa
except ModuleNotFoundError:
    VERSION = "0.0.0"
    print(
        "WARNING: AstroPhot version number not found. This is likely because you are running AstroPhot from a source directory."
    )


# meta data
__version__ = VERSION
__author__ = "Connor Stone"
__email__ = "connorstone628@gmail.com"


def run_from_terminal() -> None:
    """
    Execute AstroPhot from the command line with various options.

    This function uses the `argparse` module to parse command line arguments and execute the appropriate functionality.
    It accepts the following arguments:

    - `filename`: the path to the configuration file. Or just 'tutorial' to download tutorials.
    - `--config`: the type of configuration file being provided. One of: astrophot, galfit.
    - `-v`, `--version`: print the current AstroPhot version to screen.
    - `--log`: set the log file name for AstroPhot. Use 'none' to suppress the log file.
    - `-q`: quiet flag to stop command line output, only print to log file.
    - `--dtype`: set the float point precision. Must be one of: float64, float32.
    - `--device`: set the device for AstroPhot to use for computations. Must be one of: cpu, gpu.

    If the `filename` argument is not provided, it raises a `RuntimeError`.
    If the `filename` argument is `tutorial` or `tutorials`,
    it downloads tutorials from various URLs and saves them locally.

    This function logs messages using the `AP_config` module,
    which sets the logging output based on the `--log` and `-q` arguments.
    The `dtype` and `device` of AstroPhot can also be set using the `--dtype` and `--device` arguments, respectively.

    Returns:
        None

    """
    AP_config.ap_logger.debug("running from the terminal, not sure if it will catch me.")
    parser = argparse.ArgumentParser(
        prog="astrophot",
        description="Fast and flexible astronomical image photometry package. For the documentation go to: https://github.com/Autostronomy/AstroPhot",
        epilog="Please see the documentation or contact connor stone (connorstone628@gmail.com) for further assistance.",
    )
    parser.add_argument(
        "filename",
        nargs="?",
        metavar="configfile",
        help="the path to the configuration file. Or just 'tutorial' to download tutorials.",
    )
    parser.add_argument(
        "--config",
        type=str,
        default="astrophot",
        choices=["astrophot", "galfit"],
        metavar="format",
        help="The type of configuration file being being provided. One of: astrophot, galfit.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="print the current AstroPhot version to screen",
    )
    parser.add_argument(
        "--log",
        type=str,
        metavar="logfile.log",
        help="set the log file name for AstroPhot. use 'none' to suppress the log file.",
    )
    parser.add_argument(
        "-q",
        action="store_true",
        help="quiet flag to stop command line output, only print to log file",
    )
    parser.add_argument(
        "--dtype",
        type=str,
        choices=["float64", "float32"],
        metavar="datatype",
        help="set the float point precision. Must be one of: float64, float32",
    )
    parser.add_argument(
        "--device",
        type=str,
        choices=["cpu", "gpu"],
        metavar="device",
        help="set the device for AstroPhot to use for computations. Must be one of: cpu, gpu",
    )

    args = parser.parse_args()

    if args.log is not None:
        AP_config.set_logging_output(
            stdout=not args.q, filename=None if args.log == "none" else args.log
        )
    elif args.q:
        AP_config.set_logging_output(stdout=not args.q, filename="AstroPhot.log")

    if args.dtype is not None:
        AP_config.dtype = torch.float64 if args.dtype == "float64" else torch.float32
    if args.device is not None:
        AP_config.device = "cpu" if args.device == "cpu" else "cuda:0"

    if args.filename is None:
        raise RuntimeError(
            "Please pass a config file to astrophot. See 'astrophot --help' for more information, or go to https://Autostronomy.github.io/AstroPhot/getting_started.html"
        )
    if args.filename in ["tutorial", "tutorials"]:
        tutorials = [
            "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/GettingStarted.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/GroupModels.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/ModelZoo.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/JointModels.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/FittingMethods.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/CustomModels.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/AdvancedPSFModels.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot-tutorials/main/tutorials/ConstrainedModels.ipynb",
            "https://raw.github.com/Autostronomy/AstroPhot/main/docs/tutorials/simple_config.py",
        ]
        for url in tutorials:
            try:
                R = requests.get(url)
                with open(url[url.rfind("/") + 1 :], "w") as f:
                    f.write(R.text)
            except:
                print(
                    f"WARNING: couldn't find tutorial: {url[url.rfind('/')+1:]} check internet conection"
                )

        AP_config.ap_logger.info("collected the tutorials")
    elif args.config == "astrophot":
        basic_config(args.filename)
    elif args.config == "galfit":
        galfit_config(args.filename)
    else:
        raise ValueError(
            f"Unrecognized configuration file format {args.config}. Should be one of: astrophot, galfit"
        )



from functools import partial
from typing import Optional, Union
import io

import numpy as np
import torch

from .core_model import AstroPhot_Model
from ..image import (
    Image,
    Model_Image,
    Window,
    PSF_Image,
    Target_Image,
    Target_Image_List,
    Image,
)
from ..param import Parameter_Node, Param_Unlock, Param_SoftLimits
from ..utils.initialize import center_of_mass
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ._shared_methods import select_target
from .. import AP_config
from ..errors import InvalidTarget

__all__ = ["Component_Model"]


class Component_Model(AstroPhot_Model):
    """Component_Model(name, target, window, locked, **kwargs)

    Component_Model is a base class for models that represent single
    objects or parametric forms. It provides the basis for subclassing
    models and requires the definition of parameters, initialization,
    and model evaluation functions. This class also handles
    integration, PSF convolution, and computing the Jacobian matrix.

    Attributes:
      parameter_specs (dict): Specifications for the model parameters.
      _parameter_order (tuple): Fixed order of parameters.
      psf_mode (str): Technique and scope for PSF convolution.
      sampling_mode (str): Method for initial sampling of model. Can be one of midpoint, trapezoid, simpson. Default: midpoint
      sampling_tolerance (float): accuracy to which each pixel should be evaluated. Default: 1e-2
      integrate_mode (str): Integration scope for the model. One of none, threshold, full where threshold will select which pixels to integrate while full (in development) will integrate all pixels. Default: threshold
      integrate_max_depth (int): Maximum recursion depth when performing sub pixel integration.
      integrate_gridding (int): Amount by which to subdivide pixels when doing recursive pixel integration.
      integrate_quad_level (int): The initial quadrature level for sub pixel integration. Please always choose an odd number 3 or higher.
      softening (float): Softening length used for numerical stability and integration stability to avoid discontinuities (near R=0). Effectively has units of arcsec. Default: 1e-5
      jacobian_chunksize (int): Maximum size of parameter list before jacobian will be broken into smaller chunks.
      special_kwargs (list): Parameters which are treated specially by the model object and should not be updated directly.
      useable (bool): Indicates if the model is useable.

    Methods:
      initialize: Determine initial values for the center coordinates.
      sample: Evaluate the model on the space covered by an image object.
      jacobian: Compute the Jacobian matrix for this model.

    """

    # Specifications for the model parameters including units, value, uncertainty, limits, locked, and cyclic
    parameter_specs = {
        "center": {"units": "arcsec", "uncertainty": [0.1, 0.1]},
    }
    # Fixed order of parameters for all methods that interact with the list of parameters
    _parameter_order = ("center",)

    # Scope for PSF convolution
    psf_mode = "none"  # none, full
    # Technique for PSF convolution
    psf_convolve_mode = "fft"  # fft, direct
    # Method to use when performing subpixel shifts. bilinear set by default for stability around pixel edges, though lanczos:3 is also fairly stable, and all are stable when away from pixel edges
    psf_subpixel_shift = "bilinear"  # bilinear, lanczos:2, lanczos:3, lanczos:5, none

    # Method for initial sampling of model
    sampling_mode = (
        "midpoint"  # midpoint, trapezoid, simpsons, quad:x (where x is a positive integer)
    )

    # Level to which each pixel should be evaluated
    sampling_tolerance = 1e-2

    # Integration scope for model
    integrate_mode = "threshold"  # none, threshold

    # Maximum recursion depth when performing sub pixel integration
    integrate_max_depth = 3

    # Amount by which to subdivide pixels when doing recursive pixel integration
    integrate_gridding = 5

    # The initial quadrature level for sub pixel integration. Please always choose an odd number 3 or higher
    integrate_quad_level = 3

    # Maximum size of parameter list before jacobian will be broken into smaller chunks, this is helpful for limiting the memory requirements to build a model, lower jacobian_chunksize is slower but uses less memory
    jacobian_chunksize = 10
    image_chunksize = 1000

    # Softening length used for numerical stability and/or integration stability to avoid discontinuities (near R=0)
    softening = 1e-3

    # Parameters which are treated specially by the model object and should not be updated directly when initializing
    special_kwargs = ["parameters", "filename", "model_type"]
    track_attrs = [
        "psf_mode",
        "psf_convolve_mode",
        "psf_subpixel_shift",
        "sampling_mode",
        "sampling_tolerance",
        "integrate_mode",
        "integrate_max_depth",
        "integrate_gridding",
        "integrate_quad_level",
        "jacobian_chunksize",
        "image_chunksize",
        "softening",
    ]
    useable = False

    def __init__(self, *, name=None, **kwargs):
        self._target_identity = None
        super().__init__(name=name, **kwargs)

        self.psf = None
        self.psf_aux_image = None

        # Set any user defined attributes for the model
        for kwarg in kwargs:  # fixme move to core model?
            # Skip parameters with special behaviour
            if kwarg in self.special_kwargs:
                continue
            # Set the model parameter
            setattr(self, kwarg, kwargs[kwarg])

        # If loading from a file, get model configuration then exit __init__
        if "filename" in kwargs:
            self.load(kwargs["filename"], new_name=name)
            return

        self.parameter_specs = self.build_parameter_specs(kwargs.get("parameters", None))
        with torch.no_grad():
            self.build_parameters()
            if isinstance(kwargs.get("parameters", None), torch.Tensor):
                self.parameters.value = kwargs["parameters"]

    def set_aux_psf(self, aux_psf, add_parameters=True):
        """Set the PSF for this model as an auxiliary psf model. This psf
        model will be resampled as part of the model sampling step to
        track changes made during fitting.

        Args:
          aux_psf: The auxiliary psf model
          add_parameters: if true, the parameters of the auxiliary psf model will become model parameters for this model as well.

        """

        self._psf = aux_psf

        if add_parameters:
            self.parameters.link(aux_psf.parameters)

    @property
    def psf(self):
        if self._psf is None:
            try:
                return self.target.psf
            except AttributeError:
                return None
        return self._psf

    @psf.setter
    def psf(self, val):
        if val is None:
            self._psf = None
        elif isinstance(val, PSF_Image):
            self._psf = val
        elif isinstance(val, AstroPhot_Model):
            self.set_aux_psf(val)
        else:
            self._psf = PSF_Image(data=val, pixelscale=self.target.pixelscale)
            AP_config.ap_logger.warning(
                "Setting PSF with pixel matrix, assuming target pixelscale is the same as "
                "PSF pixelscale. To remove this warning, set PSFs as an ap.image.PSF_Image "
                "or ap.models.AstroPhot_Model object instead."
            )

    # Initialization functions
    ######################################################################
    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(
        self,
        target: Optional["Target_Image"] = None,
        parameters: Optional[Parameter_Node] = None,
        **kwargs,
    ):
        """Determine initial values for the center coordinates. This is done
        with a local center of mass search which iterates by finding
        the center of light in a window, then iteratively updates
        until the iterations move by less than a pixel.

        Args:
          target (Optional[Target_Image]): A target image object to use as a reference when setting parameter values

        """
        super().initialize(target=target, parameters=parameters)
        # Get the sub-image area corresponding to the model image
        target_area = target[self.window]

        # Use center of window if a center hasn't been set yet
        if parameters["center"].value is None:
            with (
                Param_Unlock(parameters["center"]),
                Param_SoftLimits(parameters["center"]),
            ):
                parameters["center"].value = self.window.center
        else:
            return

        if parameters["center"].locked:
            return

        # Convert center coordinates to target area array indices
        init_icenter = target_area.plane_to_pixel(parameters["center"].value)

        # Compute center of mass in window
        COM = center_of_mass(
            (
                init_icenter[1].detach().cpu().item(),
                init_icenter[0].detach().cpu().item(),
            ),
            target_area.data.detach().cpu().numpy(),
        )
        if np.any(np.array(COM) < 0) or np.any(np.array(COM) >= np.array(target_area.data.shape)):
            AP_config.ap_logger.warning("center of mass failed, using center of window")
            return
        COM = (COM[1], COM[0])
        # Convert center of mass indices to coordinates
        COM_center = target_area.pixel_to_plane(
            torch.tensor(COM, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        )

        # Set the new coordinates as the model center
        parameters["center"].value = COM_center

    # Fit loop functions
    ######################################################################
    def evaluate_model(
        self,
        X: Optional[torch.Tensor] = None,
        Y: Optional[torch.Tensor] = None,
        image: Optional[Image] = None,
        parameters: Parameter_Node = None,
        **kwargs,
    ):
        """Evaluate the model on every pixel in the given image. The
        basemodel object simply returns zeros, this function should be
        overloaded by subclasses.

        Args:
          image (Image): The image defining the set of pixels on which to evaluate the model

        """
        if X is None or Y is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]
        return torch.zeros_like(X)  # do nothing in base model

    def sample(
        self,
        image: Optional[Image] = None,
        window: Optional[Window] = None,
        parameters: Optional[Parameter_Node] = None,
    ):
        """Evaluate the model on the space covered by an image object. This
        function properly calls integration methods and PSF
        convolution. This should not be overloaded except in special
        cases.

        This function is designed to compute the model on a given
        image or within a specified window. It takes care of sub-pixel
        sampling, recursive integration for high curvature regions,
        PSF convolution, and proper alignment of the computed model
        with the original pixel grid. The final model is then added to
        the requested image.

        Args:
          image (Optional[Image]): An AstroPhot Image object (likely a Model_Image)
                                     on which to evaluate the model values. If not
                                     provided, a new Model_Image object will be created.
          window (Optional[Window]): A window within which to evaluate the model.
                                   Should only be used if a subset of the full image
                                   is needed. If not provided, the entire image will
                                   be used.

        Returns:
          Image: The image with the computed model values.

        """
        # Image on which to evaluate model
        if image is None:
            image = self.make_model_image(window=window)

        # Window within which to evaluate model
        if window is None:
            working_window = image.window.copy()
        else:
            working_window = window.copy()

        # Parameters with which to evaluate the model
        if parameters is None:
            parameters = self.parameters

        if "window" in self.psf_mode:
            raise NotImplementedError("PSF convolution in sub-window not available yet")

        if "full" in self.psf_mode:
            if isinstance(self.psf, AstroPhot_Model):
                psf = self.psf(
                    parameters=parameters[self.psf.name],
                )
            else:
                psf = self.psf
            psf_upscale = torch.round(image.pixel_length / psf.pixel_length).int()
            # Add border for psf convolution edge effects, will be cropped out later
            working_window.pad_pixel(psf.psf_border_int)
            # Make the image object to which the samples will be tracked
            working_image = Model_Image(window=working_window)
            # Sub pixel shift to align the model with the center of a pixel
            if self.psf_subpixel_shift != "none":
                pixel_center = working_image.plane_to_pixel(parameters["center"].value)
                center_shift = pixel_center - torch.round(pixel_center)
                working_image.header.pixel_shift(center_shift)
            else:
                center_shift = None

            # Evaluate the model at the current resolution
            reference, deep = self._sample_init(
                image=working_image,
                parameters=parameters,
                center=parameters["center"].value,
            )
            # If needed, super-resolve the image in areas of high curvature so pixels are properly sampled
            deep = self._sample_integrate(
                deep, reference, working_image, parameters, parameters["center"].value
            )

            # update the image with the integrated pixels
            working_image.data += deep

            # Convolve the PSF
            self._sample_convolve(working_image, center_shift, psf, self.psf_subpixel_shift)

            # Shift image back to align with original pixel grid
            if self.psf_subpixel_shift != "none":
                working_image.header.pixel_shift(-center_shift)
            # Add the sampled/integrated/convolved pixels to the requested image
            working_image = working_image.reduce(psf_upscale).crop(psf.psf_border_int)

        else:
            # Create an image to store pixel samples
            working_image = Model_Image(pixelscale=image.pixelscale, window=working_window)
            # Evaluate the model on the image
            reference, deep = self._sample_init(
                image=working_image,
                parameters=parameters,
                center=parameters["center"].value,
            )
            # Super-resolve and integrate where needed
            deep = self._sample_integrate(
                deep,
                reference,
                working_image,
                parameters,
                center=parameters["center"].value,
            )
            # Add the sampled/integrated pixels to the requested image
            working_image.data += deep

        if self.mask is not None:
            working_image.data = working_image.data * torch.logical_not(self.mask)

        image += working_image

        return image

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, tar):
        if not (tar is None or isinstance(tar, Target_Image)):
            raise InvalidTarget("AstroPhot_Model target must be a Target_Image instance.")

        # If a target image list is assigned, pick out the target appropriate for this model
        if isinstance(tar, Target_Image_List) and self._target_identity is not None:
            for subtar in tar:
                if subtar.identity == self._target_identity:
                    usetar = subtar
                    break
            else:
                raise InvalidTarget(
                    f"Could not find target in Target_Image_List with matching identity "
                    f"to {self.name}: {self._target_identity}"
                )
        else:
            usetar = tar

        self._target = usetar

        # Remember the target identity to use
        try:
            self._target_identity = self._target.identity
        except AttributeError:
            pass

    def get_state(self, save_params=True):
        """Returns a dictionary with a record of the current state of the
        model.

        Specifically, the current parameter settings and the window for
        this model. From this information it is possible for the model to
        re-build itself lated when loading from disk. Note that the target
        image is not saved, this must be reset when loading the model.

        """
        state = super().get_state()
        state["window"] = self.window.get_state()
        if save_params:
            state["parameters"] = self.parameters.get_state()
        state["target_identity"] = self._target_identity
        if isinstance(self._psf, PSF_Image) or isinstance(self._psf, AstroPhot_Model):
            state["psf"] = self._psf.get_state()
        for key in self.track_attrs:
            if getattr(self, key) != getattr(self.__class__, key):
                state[key] = getattr(self, key)
        return state

    # Extra background methods for the basemodel
    ######################################################################
    from ._model_methods import radius_metric
    from ._model_methods import angular_metric
    from ._model_methods import _sample_init
    from ._model_methods import _sample_integrate
    from ._model_methods import _sample_convolve
    from ._model_methods import _integrate_reference
    from ._model_methods import _shift_psf
    from ._model_methods import build_parameter_specs
    from ._model_methods import build_parameters
    from ._model_methods import jacobian
    from ._model_methods import _chunk_jacobian
    from ._model_methods import _chunk_image_jacobian
    from ._model_methods import load



from functools import lru_cache
from typing import Callable, Optional

import torch
import matplotlib.pyplot as plt
import numpy as np
from astropy.convolution import convolve, convolve_fft
from scipy.fft import next_fast_len
from scipy.special import roots_legendre


def fft_convolve_torch(img, psf, psf_fft=False, img_prepadded=False):
    # Ensure everything is tensor
    img = torch.as_tensor(img)
    psf = torch.as_tensor(psf)

    if img_prepadded:
        s = img.size()
    else:
        s = tuple(
            next_fast_len(int(d + (p + 1) / 2), real=True)
            for d, p in zip(img.size(), psf.size())
        )  # list(int(d + (p + 1) / 2) for d, p in zip(img.size(), psf.size()))

    img_f = torch.fft.rfft2(img, s=s)

    if not psf_fft:
        psf_f = torch.fft.rfft2(psf, s=s)
    else:
        psf_f = psf

    conv_f = img_f * psf_f
    conv = torch.fft.irfft2(conv_f, s=s)

    # Roll the tensor to correct centering and crop to original image size
    return torch.roll(
        conv,
        shifts=(-int((psf.size()[0] - 1) / 2), -int((psf.size()[1] - 1) / 2)),
        dims=(0, 1),
    )[: img.size()[0], : img.size()[1]]


def fft_convolve_multi_torch(
    img, kernels, kernel_fft=False, img_prepadded=False, dtype=None, device=None
):
    # Ensure everything is tensor
    img = torch.as_tensor(img, dtype=dtype, device=device)
    for k in range(len(kernels)):
        kernels[k] = torch.as_tensor(kernels[k], dtype=dtype, device=device)

    if img_prepadded:
        s = img.size()
    else:
        s = list(int(d + (p + 1) / 2) for d, p in zip(img.size(), kernels[0].size()))

    img_f = torch.fft.rfft2(img, s=s)

    if not kernel_fft:
        kernels_f = list(torch.fft.rfft2(kernel, s=s) for kernel in kernels)
    else:
        psf_f = psf

    conv_f = img_f

    for kernel_f in kernels_f:
        conv_f *= kernel_f

    conv = torch.fft.irfft2(conv_f, s=s)

    # Roll the tensor to correct centering and crop to original image size
    return torch.roll(
        conv,
        shifts=(
            -int((sum(kernel.size()[0] for kernel in kernels) - 1) / 2),
            -int((sum(kernel.size()[1] for kernel in kernels) - 1) / 2),
        ),
        dims=(0, 1),
    )[: img.size()[0], : img.size()[1]]


def displacement_spacing(N, dtype=torch.float64, device="cpu"):
    return torch.linspace(
        -(N - 1) / (2 * N), (N - 1) / (2 * N), N, dtype=dtype, device=device
    )


def displacement_grid(Nx, Ny, pixelscale=None, dtype=torch.float64, device="cpu"):
    px = displacement_spacing(Nx, dtype=dtype, device=device)
    py = displacement_spacing(Ny, dtype=dtype, device=device)
    PX, PY = torch.meshgrid(px, py, indexing="xy")
    return (pixelscale @ torch.stack((PX, PY)).view(2, -1)).reshape((2, *PX.shape))


@lru_cache(maxsize=32)
def quad_table(n, p, dtype, device):
    """
    from: https://pomax.github.io/bezierinfo/legendre-gauss.html
    """
    abscissa, weights = roots_legendre(n)

    w = torch.tensor(weights, dtype=dtype, device=device)
    a = torch.tensor(abscissa, dtype=dtype, device=device)
    X, Y = torch.meshgrid(a, a, indexing="xy")

    W = torch.outer(w, w) / 4.0

    X, Y = p @ (torch.stack((X, Y)).view(2, -1) / 2.0)

    return X, Y, W.reshape(-1)


def single_quad_integrate(
    X, Y, image_header, eval_brightness, eval_parameters, dtype, device, quad_level=3
):

    # collect gaussian quadrature weights
    abscissaX, abscissaY, weight = quad_table(
        quad_level, image_header.pixelscale, dtype, device
    )
    # Specify coordinates at which to evaluate function
    Xs = torch.repeat_interleave(X[..., None], quad_level ** 2, -1) + abscissaX
    Ys = torch.repeat_interleave(Y[..., None], quad_level ** 2, -1) + abscissaY

    # Evaluate the model at the quadrature points
    res = eval_brightness(
        X=Xs,
        Y=Ys,
        image=image_header,
        parameters=eval_parameters,
    )

    # Reference flux for pixel is simply the mean of the evaluations
    ref = res[..., (quad_level**2) // 2] #res.mean(axis=-1) # # alternative, use midpoint
    
    # Apply the weights and reduce to original pixel space
    res = (res * weight).sum(axis=-1)

    return res, ref

def grid_integrate(
    X,
    Y,
    image_header,
    eval_brightness,
    eval_parameters,
    dtype,
    device,
    quad_level=3,
    gridding=5,
    _current_depth=1,
    max_depth=2,
    reference=None,
):
    """The grid_integrate function performs adaptive quadrature
    integration over a given pixel grid, offering precision control
    where it is needed most.

    Args:
      X (torch.Tensor): A 2D tensor representing the x-coordinates of the grid on which the function will be integrated.
      Y (torch.Tensor): A 2D tensor representing the y-coordinates of the grid on which the function will be integrated.
      image_header (ImageHeader): An object containing meta-information about the image.
      eval_brightness (callable): A function that evaluates the brightness at each grid point. This function should be compatible with PyTorch tensor operations.
      eval_parameters (Parameter_Group): An object containing parameters that are passed to the eval_brightness function.
      dtype (torch.dtype): The data type of the output tensor. The dtype argument should be a valid PyTorch data type.
      device (torch.device): The device on which to perform the computations. The device argument should be a valid PyTorch device.
      quad_level (int, optional): The initial level of quadrature used in the integration. Defaults to 3.
      gridding (int, optional): The factor by which the grid is subdivided when the integration error for a pixel is above the allowed threshold. Defaults to 5.
      _current_depth (int, optional): The current depth level of the grid subdivision. Used for recursive calls to the function. Defaults to 1.
      max_depth (int, optional): The maximum depth level of grid subdivision. Once this level is reached, no further subdivision is performed. Defaults to 2.
      reference (torch.Tensor or None, optional): A scalar value that represents the allowed threshold for the integration error. 

    Returns:
      torch.Tensor: A tensor of the same shape as X and Y that represents the result of the integration on the grid.

    This function operates by first performing a quadrature
    integration over the given pixels. If the maximum depth level has
    been reached, it simply returns the result. Otherwise, it
    calculates the integration error for each pixel and selects those
    that have an error above the allowed threshold. For pixels that
    have low error, the result is set as computed. For those with high
    error, it sets up a finer sampling grid and recursively evaluates
    the quadrature integration on it. Finally, it integrates the
    results from the finer sampling grid back to the current
    resolution.

    """
    # perform quadrature integration on the given pixels
    res, ref = single_quad_integrate(
        X,
        Y,
        image_header,
        eval_brightness,
        eval_parameters,
        dtype,
        device,
        quad_level=quad_level,
    )

    # if the max depth is reached, simply return the integrated pixels
    if _current_depth >= max_depth:
        return res

    # Begin integral
    integral = torch.zeros_like(X)

    # Select pixels which have errors above the allowed threshold
    select = torch.abs((res - ref)) > reference

    # For pixels with low error, set the results as computed
    integral[torch.logical_not(select)] = res[torch.logical_not(select)]

    # Set up sub-gridding to super resolve problem pixels
    stepx, stepy = displacement_grid(
        gridding, gridding, image_header.pixelscale, dtype, device
    )
    # Write out the coordinates for the super resolved pixels
    subgridX = torch.repeat_interleave(
        X[select].unsqueeze(-1), gridding ** 2, -1
    ) + stepx.reshape(-1)
    subgridY = torch.repeat_interleave(
        Y[select].unsqueeze(-1), gridding ** 2, -1
    ) + stepy.reshape(-1)

    # Recursively evaluate the quadrature integration on the finer sampling grid
    subgridres = grid_integrate(
        subgridX,
        subgridY,
        image_header.rescale_pixel(1/gridding),
        eval_brightness,
        eval_parameters,
        dtype,
        device,
        quad_level=quad_level,
        gridding=gridding,
        _current_depth=_current_depth+1,
        max_depth=max_depth,
        reference=reference * gridding**2,        
    )

    # Integrate the finer sampling grid back to current resolution
    integral[select] = subgridres.sum(axis=(-1,))

    return integral



from .core_model import *
from .model_object import *
from .galaxy_model_object import *
from .ray_model import *
from .sersic_model import *
from .group_model_object import *
from .sky_model_object import *
from .flatsky_model import *
from .planesky_model import *
from .gaussian_model import *
from .spline_model import *
from .relspline_model import *
from .psf_model_object import *
from .pixelated_psf_model import *
from .eigen_psf import *
from .superellipse_model import *
from .edgeon_model import *
from .exponential_model import *
from .foureirellipse_model import *
from .wedge_model import *
from .warp_model import *
from .moffat_model import *
from .nuker_model import *
from .zernike_model import *
from .airy_psf import *
from .point_source import *
from .group_psf_model import *



import numpy as np

from .center import Lanczos_peak, center_of_mass, GaussianDensity_Peak
from ..interpolate import shift_Lanczos_np, point_Lanczos


def gaussian_psf(sigma, img_width, pixelscale, upsample=4):
    assert img_width % 2 == 1, "psf images should have an odd shape"

    # Number of super sampled pixels
    N = img_width * upsample
    # Grid of centered pixel locations
    XX, YY = np.meshgrid(
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
    )
    # Evaluate the Gaussian at each pixel
    ZZ = np.exp(-0.5 * (XX ** 2 + YY ** 2) / sigma ** 2)

    # Reduce the super-sampling back to native resolution
    ZZ = ZZ.reshape(img_width, upsample, img_width, upsample).sum(axis=(1, 3)) / (
        upsample ** 2
    )

    # Normalize the PSF
    return ZZ / np.sum(ZZ)


def moffat_psf(n, Rd, img_width, pixelscale, upsample=4):
    assert img_width % 2 == 1, "psf images should have an odd shape"

    # Number of super sampled pixels
    N = img_width * upsample
    # Grid of centered pixel locations
    XX, YY = np.meshgrid(
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
        np.linspace(
            -(N - 1) * pixelscale / (2 * upsample),
            (N - 1) * pixelscale / (2 * upsample),
            N,
        ),
    )
    # Evaluate the Moffat at each pixel
    ZZ = 1.0 / (1.0 + (XX ** 2 + YY ** 2) / (Rd ** 2)) ** n

    # Reduce the super-sampling back to native resolution
    ZZ = ZZ.reshape(img_width, upsample, img_width, upsample).sum(axis=(1, 3)) / (
        upsample ** 2
    )

    # Normalize the PSF
    return ZZ / np.sum(ZZ)


def construct_psf(
    stars, image, sky_est, size=51, mask=None, keep_init=False, Lanczos_scale=3
):
    """Given a list of initial guesses for star center locations, finds
    the interpolated flux peak, re-centers the stars such that they
    are exactly on a pixel center, then median stacks the normalized
    stars to determine an average PSF.

    Note that all coordinates in this function are pixel
    coordinates. That is, the image[0][0] pixel is at location (0,0)
    and the image[2][7] pixel is at location (2,7) in this coordinate
    system.
    """
    size += 1 - (size % 2)
    star_centers = []
    # determine exact (sub-pixel) center for each star

    for star in stars:
        if keep_init:
            star_centers = list(np.array(s) for s in stars)
            break
        try:
            peak = GaussianDensity_Peak(star, image)
        except Exception as e:
            AP_config.ap_logger.warning("issue finding star center")
            AP_config.ap_logger.warning(e)
            AP_config.ap_logger.warning("skipping")
            continue
        pixel_cen = np.round(peak)
        if (
            pixel_cen[0] < ((size - 1) / 2)
            or pixel_cen[0] > (image.shape[1] - ((size - 1) / 2) - 1)
            or pixel_cen[1] < ((size - 1) / 2)
            or pixel_cen[1] > (image.shape[0] - ((size - 1) / 2) - 1)
        ):
            AP_config.ap_logger.debug("skipping star near edge at: {peak}")
            continue
        star_centers.append(peak)

    stacking = []
    # Extract the star from the image, and shift to align exactly with pixel grid
    for star in star_centers:
        center = np.round(star)
        border = int((size - 1) / 2 + Lanczos_scale)
        I = image[
            int(center[1] - border) : int(center[1] + border + 1),
            int(center[0] - border) : int(center[0] + border + 1),
        ]
        shift = center - star
        I = shift_Lanczos_np(I - sky_est, shift[0], shift[1], scale=Lanczos_scale)
        I = I[Lanczos_scale:-Lanczos_scale, Lanczos_scale:-Lanczos_scale]
        border = (size - 1) / 2
        if mask is not None:
            I[
                mask[
                    int(center[1] - border) : int(center[1] + border + 1),
                    int(center[0] - border) : int(center[0] + border + 1),
                ]
            ] = np.nan
        # Add the normalized star image to the list
        stacking.append(I / np.sum(I))

    # Median stack the pixel images
    stacked_psf = np.nanmedian(stacking, axis=0)
    stacked_psf[stacked_psf < 0] = 0

    return stacked_psf / np.sum(stacked_psf)



import os
root_path = os.path.split(os.path.split(os.path.split(__file__)[0])[0])[0]

with open(os.path.join(root_path, "VERSION"), "r") as f:
    version = f.read().strip()

print(root_path)

print(version)

# Update CITATION.cff
######################################################################
with open(os.path.join(root_path, "CITATION.cff"), "r") as f:
    CITATION = f.readlines()

with open(os.path.join(root_path, "CITATION.cff"), "w") as f:
    for line in CITATION:
        if line.startswith("version:"):
            f.write(f"version: {version}\n")
        else:
            f.write(line)

# Update docs/conf.py
######################################################################
with open(os.path.join(root_path, "docs", "conf.py"), "r") as f:
    CONF = f.readlines()

with open(os.path.join(root_path, "docs", "conf.py"), "w") as f:
    for line in CONF:
        if line.startswith("version ="):
            f.write(f"version = \"{version[:version.rfind('.')]}\"\n")
        elif line.startswith("release ="):
            f.write(f"release = \"{version}\"\n")
        else:
            f.write(line)

# Update astrophot/__init__.py
######################################################################
with open(os.path.join(root_path, "astrophot", "__init__.py"), "r") as f:
    INIT = f.readlines()

with open(os.path.join(root_path, "astrophot", "__init__.py"), "w") as f:
    for line in INIT:
        if line.startswith("__version__ ="):
            f.write(f"__version__ = \"{version}\"\n")
        else:
            f.write(line)



import unittest
import astrophot as ap
import torch
import numpy as np
from utils import make_basic_sersic, make_basic_gaussian, make_basic_gaussian_psf
#torch.autograd.set_detect_anomaly(True)
######################################################################
# Model Objects
######################################################################


class TestModel(unittest.TestCase):
    def test_AstroPhot_Model(self):

        model = ap.models.AstroPhot_Model(name = "test model")

        self.assertIsNone(model.target, "model should not have a target at this point")

        target = ap.image.Target_Image(data=torch.zeros((16, 32)), pixelscale=1.0)

        model.target = target

        model.window = target.window

        model.locked = True
        model.locked = False

        state = model.get_state()

    def test_initialize_does_not_recurse(self):
        "Test case for error where missing parameter name triggered print that triggered missing parameter name ..."
        target = make_basic_sersic()
        model = ap.models.AstroPhot_Model(
            name="test model",
            model_type="sersic galaxy model",
            target=target,
        )
        # Define a function that accesses a parameter that doesn't exist
        def calc(params):
            return params["A"].value

        model["center"].value = calc

        with self.assertRaises(KeyError) as context:
            model.initialize()

    def test_basic_model_methods(self):

        target = make_basic_sersic()
        model = ap.models.AstroPhot_Model(
            name="test sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            target=target,
        )
        rep = model.parameters.vector_representation()
        nat = model.parameters.vector_values()
        self.assertTrue(torch.all(torch.isclose(rep, model.parameters.vector_transform_val_to_rep(nat))), "transform should map between parameter natural and representation")
        self.assertTrue(torch.all(torch.isclose(nat, model.parameters.vector_transform_rep_to_val(rep))), "transform should map between parameter representation and natural")

    def test_model_sampling_modes(self):

        target = make_basic_sersic(100,100)
        model = ap.models.AstroPhot_Model(
            name="test sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            target=target,
        )
        res = model()
        model.sampling_mode = "trapezoid"
        res = model()
        model.sampling_mode = "simpsons"
        res = model()
        model.sampling_mode = "quad:3"
        res = model()
        model.integrate_mode = "none"
        res = model()
        model.integrate_mode = "should raise"
        self.assertRaises(ap.errors.SpecificationConflict, model)
        model.integrate_mode = "none"
        model.sampling_mode = "should raise"
        self.assertRaises(ap.errors.SpecificationConflict, model)     
        model.sampling_mode = "midpoint"   

        # test PSF modes
        model.psf = np.array([[0.05, 0.1, 0.05], [0.1, 0.4, 0.1], [0.05, 0.1, 0.05]])
        model.integrate_mode = "none"
        model.psf_mode = "full"
        model.psf_convolve_mode = "direct"
        res = model()
        model.psf_convolve_mode = "fft"
        res = model()

    def test_model_creation(self):
        np.random.seed(12345)
        shape = (10, 15)
        tar = ap.image.Target_Image(
            data=np.random.normal(loc=0, scale=1.4, size=shape),
            pixelscale=0.8,
            variance=np.ones(shape) * (1.4 ** 2),
            psf=np.array([[0.05, 0.1, 0.05], [0.1, 0.4, 0.1], [0.05, 0.1, 0.05]]),
        )

        mod = ap.models.Component_Model(
            name="base model",
            target=tar,
            parameters={"center": {"value": [5, 5], "locked": True}},
        )

        mod.initialize()

        self.assertFalse(mod.locked, "default model should not be locked")

        self.assertTrue(
            torch.all(mod().data == 0), "Component_Model model_image should be zeros"
        )
    
    def test_mask(self):

        target = make_basic_sersic()
        mask = torch.zeros_like(target.data)
        mask[10,13] = 1
        model = ap.models.AstroPhot_Model(
            name="test sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            target=target,
            mask = mask
        )

        sample = model()
        self.assertEqual(sample.data[10,13].item(), 0., "masked values should be zero")
        self.assertNotEqual(sample.data[11,12].item(), 0., "unmasked values should NOT be zero")

    def test_model_errors(self):

        # Invalid name
        self.assertRaises(ap.errors.NameNotAllowed, ap.models.AstroPhot_Model, name = "my|model")

        # Target that is not a target image
        arr = torch.zeros((10, 15))
        target = ap.image.Image(
            data = arr, pixelscale=1.0, zeropoint=1.0, origin=torch.zeros(2)
        )

        with self.assertRaises(ap.errors.InvalidTarget):
            model = ap.models.AstroPhot_Model(
                name="test model",
                model_type="sersic galaxy model",
                target=target,
            )

        # model that doesnt exist
        target = make_basic_sersic()
        with self.assertRaises(ap.errors.UnrecognizedModel):
            model = ap.models.AstroPhot_Model(
                name="test model",
                model_type="sersic gaaxy model",
                target=target,
            )

        # invalid window
        with self.assertRaises(ap.errors.InvalidWindow):
            model = ap.models.AstroPhot_Model(
                name="test model",
                model_type="sersic galaxy model",
                target=target,
                window = (1,2,3),
            )
            
        


class TestAllModelBasics(unittest.TestCase):
    def test_all_model_sample(self):

        target = make_basic_sersic()
        for model_type in ap.models.Component_Model.List_Model_Names(useable=True):
            print(model_type)
            MODEL = ap.models.AstroPhot_Model(
                name="test model",
                model_type=model_type,
                target=target,
            )
            MODEL.initialize()
            for P in MODEL.parameter_order:
                self.assertIsNotNone(
                    MODEL[P].value,
                    f"Model type {model_type} parameter {P} should not be None after initialization",
                )
            img = MODEL()
            self.assertTrue(
                torch.all(torch.isfinite(img.data)),
                "Model should evaluate a real number for the full image",
            )
            self.assertIsInstance(str(MODEL), str, "String representation should return string")
            self.assertIsInstance(repr(MODEL), str, "Repr should return string")
        
class TestSersic(unittest.TestCase):
    def test_sersic_creation(self):
        np.random.seed(12345)
        N = 50
        Width = 20
        shape = (N + 10, N)
        true_params = [2, 5, 10, -3, 5, 0.7, np.pi / 4]
        IXX, IYY = np.meshgrid(
            np.linspace(-Width, Width, shape[1]), np.linspace(-Width, Width, shape[0])
        )
        QPAXX, QPAYY = ap.utils.conversions.coordinates.Axis_Ratio_Cartesian_np(
            true_params[5], IXX - true_params[3], IYY - true_params[4], true_params[6]
        )
        Z0 = ap.utils.parametric_profiles.sersic_np(
            np.sqrt(QPAXX ** 2 + QPAYY ** 2),
            true_params[0],
            true_params[1],
            true_params[2],
        ) + np.random.normal(loc=0, scale=0.1, size=shape)
        tar = ap.image.Target_Image(
            data=Z0,
            pixelscale=0.8,
            variance=np.ones(Z0.shape) * (0.1 ** 2),
        )

        mod = ap.models.Sersic_Galaxy(
            name="sersic model",
            target=tar,
            parameters={"center": [-3.2 + N / 2, 5.1 + (N + 10) / 2]},
        )

        self.assertFalse(mod.locked, "default model should not be locked")

        mod.initialize()

    def test_sersic_save_load(self):

        target = make_basic_sersic()
        psf = make_basic_gaussian_psf()
        model = ap.models.AstroPhot_Model(
            name="test sersic",
            model_type="sersic galaxy model",
            parameters={
                "center": [20, 20],
                "PA": 60 * np.pi / 180,
                "q": 0.5,
                "n": 2,
                "Re": 5,
                "Ie": 1,
            },
            psf = psf,
            psf_mode = "full",
            target=target,
        )

        model.initialize()
        model.save("test_AstroPhot_sersic.yaml")
        model2 = ap.models.AstroPhot_Model(
            name="load model",
            filename="test_AstroPhot_sersic.yaml",
        )

        for P in model.parameter_order:
            self.assertAlmostEqual(
                model[P].value.detach().cpu().tolist(),
                model2[P].value.detach().cpu().tolist(),
                msg="loaded model should have same parameters",
            )


if __name__ == "__main__":
    unittest.main()



# Apply a different optimizer iteratively
from typing import Dict, Any, Sequence, Union
import os
from time import time
from copy import copy
import random

import numpy as np
import torch
from scipy.optimize import minimize
from scipy.special import gammainc

from .base import BaseOptimizer
from ..models import AstroPhot_Model
from .lm import LM
from ..param import Param_Mask
from .. import AP_config

__all__ = ["Iter", "Iter_LM"]


class Iter(BaseOptimizer):
    """Optimizer wrapper that performs optimization iteratively.

    This optimizer applies a different optimizer to a group model iteratively.
    It can be used for complex fits or when the number of models to fit is too large to fit in memory.

    Args:
        model: An `AstroPhot_Model` object to perform optimization on.
        method: The optimizer class to apply at each iteration step.
        initial_state: Optional initial state for optimization, defaults to None.
        max_iter: Maximum number of iterations, defaults to 100.
        method_kwargs: Keyword arguments to pass to `method`.
        **kwargs: Additional keyword arguments.

    Attributes:
        ndf: Degrees of freedom of the data.
        method: The optimizer class to apply at each iteration step. Default: Levenberg-Marquardt
        method_kwargs: Keyword arguments to pass to `method`.
        iteration: The number of iterations performed.
        lambda_history: A list of the states at each iteration step.
        loss_history: A list of the losses at each iteration step
    """

    def __init__(
        self,
        model: AstroPhot_Model,
        method: BaseOptimizer = LM,
        initial_state: np.ndarray = None,
        max_iter: int = 100,
        method_kwargs: Dict[str, Any] = {},
        **kwargs: Dict[str, Any],
    ) -> None:
        super().__init__(model, initial_state, max_iter=max_iter, **kwargs)

        self.method = method
        self.method_kwargs = method_kwargs
        if "relative_tolerance" not in method_kwargs and isinstance(method, LM):
            # Lower tolerance since it's not worth fine tuning a model when its neighbors will be shifting soon anyway
            self.method_kwargs["relative_tolerance"] = 1e-3
            self.method_kwargs["max_iter"] = 15
        #          # pixels      # parameters
        self.ndf = self.model.target[self.model.window].flatten("data").size(0) - len(
            self.current_state
        )
        if self.model.target.has_mask:
            # subtract masked pixels from degrees of freedom
            self.ndf -= torch.sum(
                self.model.target[self.model.window].flatten("mask")
            ).item()

    def sub_step(self, model: "AstroPhot_Model") -> None:
        """
        Perform optimization for a single model.

        Args:
            model: The model to perform optimization on.
        """
        self.Y -= model()
        initial_target = model.target
        model.target = model.target[model.window] - self.Y[model.window]
        res = self.method(model, **self.method_kwargs).fit()
        self.Y += model()
        if self.verbose > 1:
            AP_config.ap_logger.info(res.message)
        model.target = initial_target
        
    def step(self) -> None:
        """
        Perform a single iteration of optimization.
        """
        if self.verbose > 0:
            AP_config.ap_logger.info("--------iter-------")

        # Fit each model individually
        for model in self.model.models.values():
            if self.verbose > 0:
                AP_config.ap_logger.info(model.name)
            self.sub_step(model)
        # Update the current state
        self.current_state = self.model.parameters.vector_representation()

        # Update the loss value
        with torch.no_grad():
            if self.verbose > 0:
                AP_config.ap_logger.info("Update Chi^2 with new parameters")
            self.Y = self.model(
                parameters=self.current_state,
                as_representation=True,
            )
            D = self.model.target[self.model.window].flatten("data")
            V = (
                self.model.target[self.model.window].flatten("variance")
                if self.model.target.has_variance
                else 1.0
            )
            if self.model.target.has_mask:
                M = self.model.target[self.model.window].flatten("mask")
                loss = (
                    torch.sum(
                        (((D - self.Y.flatten("data")) ** 2) / V)[torch.logical_not(M)]
                    )
                    / self.ndf
                )
            else:
                loss = torch.sum(((D - self.Y.flatten("data")) ** 2 / V)) / self.ndf
        if self.verbose > 0:
            AP_config.ap_logger.info(f"Loss: {loss.item()}")
        self.lambda_history.append(np.copy((self.current_state).detach().cpu().numpy()))
        self.loss_history.append(loss.item())

        # Test for convergence
        if self.iteration >= 2 and (
            (-self.relative_tolerance * 1e-3)
            < ((self.loss_history[-2] - self.loss_history[-1]) / self.loss_history[-1])
            < (self.relative_tolerance / 10)
        ):
            self._count_finish += 1
        else:
            self._count_finish = 0

        self.iteration += 1

    def fit(self) -> "BaseOptimizer":
        """
        Fit the models to the target.


        """

        self.iteration = 0
        self.Y = self.model(parameters=self.current_state, as_representation=True)
        start_fit = time()
        try:
            while True:
                self.step()
                if self.save_steps is not None:
                    self.model.save(
                        os.path.join(
                            self.save_steps,
                            f"{self.model.name}_Iteration_{self.iteration:03d}.yaml",
                        )
                    )
                if self.iteration > 2 and self._count_finish >= 2:
                    self.message = self.message + "success"
                    break
                elif self.iteration >= self.max_iter:
                    self.message = (
                        self.message + f"fail max iterations reached: {self.iteration}"
                    )
                    break

        except KeyboardInterrupt:
            self.message = self.message + "fail interrupted"

        self.model.parameters.vector_set_representation(self.res())
        if self.verbose > 1:
            AP_config.ap_logger.info(
                f"Iter Fitting complete in {time() - start_fit} sec with message: {self.message}"
            )

        return self


class Iter_LM(BaseOptimizer):
    """Optimization wrapper that call LM optimizer on subsets of variables.

    Iter_LM takes the full set of parameters for a model and breaks
    them down into chunks as specified by the user. It then calls
    Levenberg-Marquardt optimization on the subset of parameters, and
    iterates through all subsets until every parameter has been
    optimized. It cycles through these chunks until convergence. This
    method is very powerful in situations where the full optimization
    problem cannot fit in memory, or where the optimization problem is
    too complex to tackle as a single large problem. In full LM
    optimization a single problematic parameter can ripple into issues
    with every other parameter, so breaking the problem down can
    sometimes make an otherwise intractable problem easier. For small
    problems with only a few models, it is likely better to optimize
    the full problem with LM as, when it works, LM is faster than the
    Iter_LM method.

    Args:
      chunks (Union[int, tuple]): Specify how to break down the model parameters. If an integer, at each iteration the algorithm will break the parameters into groups of that size. If a tuple, should be a tuple of tuples of strings which give an explicit pairing of parameters to optimize, note that it is allowed to have variable size chunks this way. Default: 50
      method (str): How to iterate through the chunks. Should be one of: random, sequential. Default: random
    """

    def __init__(
        self,
        model: "AstroPhot_Model",
        initial_state: Sequence = None,
        chunks: Union[int, tuple] = 50,
        max_iter: int = 100,
        method: str = "random",
        LM_kwargs: dict = {},
        **kwargs: Dict[str, Any],
    ) -> None:
        super().__init__(model, initial_state, max_iter=max_iter, **kwargs)

        self.chunks = chunks
        self.method = method
        self.LM_kwargs = LM_kwargs

        #          # pixels      # parameters
        self.ndf = self.model.target[self.model.window].flatten("data").numel() - len(
            self.current_state
        )
        if self.model.target.has_mask:
            # subtract masked pixels from degrees of freedom
            self.ndf -= torch.sum(
                self.model.target[self.model.window].flatten("mask")
            ).item()

    def step(self):
        # These store the chunking information depending on which chunk mode is selected
        param_ids = list(self.model.parameters.vector_identities())
        init_param_ids = list(self.model.parameters.vector_identities())
        _chunk_index = 0
        _chunk_choices = None
        res = None

        if self.verbose > 0:
            AP_config.ap_logger.info("--------iter-------")

        # Loop through all the chunks
        while True:
            chunk = torch.zeros(len(init_param_ids), dtype = torch.bool, device = AP_config.ap_device)
            if isinstance(self.chunks, int):
                if len(param_ids) == 0:
                    break
                if self.method == "random":
                    # Draw a random chunk of ids
                    for pid in random.sample(param_ids, min(len(param_ids), self.chunks)):
                        chunk[init_param_ids.index(pid)] = True
                else:
                    # Draw the next chunk of ids
                    for pid in param_ids[: self.chunks]:
                        chunk[init_param_ids.index(pid)] = True
                # Remove the selected ids from the list
                for p in np.array(init_param_ids)[chunk.detach().cpu().numpy()]:
                    param_ids.pop(param_ids.index(p))
            elif isinstance(self.chunks, (tuple, list)):
                if _chunk_choices is None:
                    # Make a list of the chunks as given explicitly
                    _chunk_choices = list(range(len(self.chunks)))
                if self.method == "random":
                    if len(_chunk_choices) == 0:
                        break
                    # Select a random chunk from the given groups
                    sub_index = random.choice(_chunk_choices)
                    _chunk_choices.pop(_chunk_choices.index(sub_index))
                    for pid in self.chunks[sub_index]:
                        chunk[param_ids.index(pid)] = True
                else:
                    if _chunk_index >= len(self.chunks):
                        break
                    # Select the next chunk in order
                    for pid in self.chunks[_chunk_index]:
                        chunk[param_ids.index(pid)] = True
                    _chunk_index += 1
            else:
                raise ValueError(
                    "Unrecognized chunks value, should be one of int, tuple. not: {type(self.chunks)}"
                )
            if self.verbose > 1:
                AP_config.ap_logger.info(str(chunk))
            del res
            with Param_Mask(self.model.parameters, chunk):
                res = LM(
                    self.model,
                    ndf=self.ndf,
                    **self.LM_kwargs,
                ).fit()
            if self.verbose > 0:
                AP_config.ap_logger.info(f"chunk loss: {res.res_loss()}")
            if self.verbose > 1:
                AP_config.ap_logger.info(f"chunk message: {res.message}")

        self.loss_history.append(res.res_loss())
        self.lambda_history.append(
            self.model.parameters.vector_representation()
            .detach()
            .cpu()
            .numpy()
        )
        if self.verbose > 0:
            AP_config.ap_logger.info(f"Loss: {self.loss_history[-1]}")

        # test for convergence
        if self.iteration >= 2 and (
            (-self.relative_tolerance * 1e-3)
            < ((self.loss_history[-2] - self.loss_history[-1]) / self.loss_history[-1])
            < (self.relative_tolerance / 10)
        ):
            self._count_finish += 1
        else:
            self._count_finish = 0

        self.iteration += 1

    def fit(self):
        self.iteration = 0

        start_fit = time()
        try:
            while True:
                self.step()
                if self.save_steps is not None:
                    self.model.save(
                        os.path.join(
                            self.save_steps,
                            f"{self.model.name}_Iteration_{self.iteration:03d}.yaml",
                        )
                    )
                if self.iteration > 2 and self._count_finish >= 2:
                    self.message = self.message + "success"
                    break
                elif self.iteration >= self.max_iter:
                    self.message = (
                        self.message + f"fail max iterations reached: {self.iteration}"
                    )
                    break

        except KeyboardInterrupt:
            self.message = self.message + "fail interrupted"

        self.model.parameters.vector_set_representation(self.res())
        if self.verbose > 1:
            AP_config.ap_logger.info(
                f"Iter Fitting complete in {time() - start_fit} sec with message: {self.message}"
            )

        return self



import numpy as np
import torch

from .galaxy_model_object import Galaxy_Model
from ..param import Parameter_Node
from ..utils.interpolate import cubic_spline_torch
from ..utils.conversions.coordinates import Axis_Ratio_Cartesian
from ..utils.decorators import default_internal

__all__ = ["Wedge_Galaxy"]


class Wedge_Galaxy(Galaxy_Model):
    """Variant of the ray model where no smooth transition is performed
    between regions as a function of theta, instead there is a sharp
    trnasition boundary. This may be desireable as it cleanly
    separates where the pixel information is going. Due to the sharp
    transition though, it may cause unusual behaviour when fitting. If
    problems occur, try fitting a ray model first then fix the center,
    PA, and q and then fit the wedge model. Essentially this breaks
    down the structure fitting and the light profile fitting into two
    steps. The wedge model, like the ray model, defines no extra
    parameters, however a new option can be supplied on instantiation
    of the wedge model which is "wedges" or the number of wedges in
    the model.

    """

    model_type = f"wedge {Galaxy_Model.model_type}"
    special_kwargs = Galaxy_Model.special_kwargs + ["wedges"]
    wedges = 2
    track_attrs = Galaxy_Model.track_attrs + ["wedges"]
    useable = False

    def __init__(self, *args, **kwargs):
        self.symmetric_wedges = True
        super().__init__(*args, **kwargs)
        self.wedges = kwargs.get("wedges", 2)

    @default_internal
    def polar_model(self, R, T, image=None, parameters=None):
        model = torch.zeros_like(R)
        if self.wedges % 2 == 0 and self.symmetric_wedges:
            for w in range(self.wedges):
                angles = (T - (w * np.pi / self.wedges)) % np.pi
                indices = torch.logical_or(
                    angles < (np.pi / (2 * self.wedges)),
                    angles >= (np.pi * (1 - 1 / (2 * self.wedges))),
                )
                model[indices] += self.iradial_model(w, R[indices], image, parameters)
        elif self.wedges % 2 == 1 and self.symmetric_wedges:
            for w in range(self.wedges):
                angles = (T - (w * np.pi / self.wedges)) % (2 * np.pi)
                indices = torch.logical_or(
                    angles < (np.pi / (2 * self.wedges)),
                    angles >= (np.pi * (2 - 1 / (2 * self.wedges))),
                )
                model[indices] += self.iradial_model(w, R[indices], image, parameters)
                angles = (T - (np.pi + w * np.pi / self.wedges)) % (2 * np.pi)
                indices = torch.logical_or(
                    angles < (np.pi / (2 * self.wedges)),
                    angles >= (np.pi * (2 - 1 / (2 * self.wedges))),
                )
                model[indices] += self.iradial_model(w, R[indices], image, parameters)
        else:
            for w in range(self.wedges):
                angles = (T - (w * 2 * np.pi / self.wedges)) % (2 * np.pi)
                indices = torch.logical_or(
                    angles < (np.pi / self.wedges),
                    angles >= (np.pi * (2 - 1 / self.wedges)),
                )
                model[indices] += self.iradial_model(w, R[indices], image, parameters)
        return model

    @default_internal
    def evaluate_model(self, X=None, Y=None, image=None, parameters=None, **kwargs):
        if X is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]
        XX, YY = self.transform_coordinates(X, Y, image, parameters)

        return self.polar_model(
            self.radius_metric(XX, YY, image=image, parameters=parameters),
            self.angular_metric(XX, YY, image=image, parameters=parameters),
            image=image,
            parameters=parameters,
        )



from collections import OrderedDict
from abc import ABC, abstractmethod
from ..errors import InvalidParameter

__all__ = ["Node"]

class Node(ABC):
    """Base node object in the Directed Acyclic Graph (DAG).

    The base Node object handles storing the DAG nodes and links
    between them. An important part of the DAG system is to be able to
    find all the leaf nodes, which is done using the `flat` function.

    Args:
      name (str): The name of the node, this should identify it uniquely in the local context it will be used in.
      locked (bool): Records if the node is locked, this is relevant for some other operations which only act on unlocked nodes.
      link (tuple[Node]): A tuple of node objects which this node will be linked to on initialization.

    """
    global_unlock = False
    
    def __init__(self, name, **kwargs):
        if ":" in name:
            raise ValueError(f"Node names must not have ':' character. Cannot use name: {name}")
        self.name = name
        self.nodes = OrderedDict()
        if "state" in kwargs:
            self.set_state(kwargs["state"])
            return
        if "link" in kwargs:
            self.link(*kwargs["link"])
        self.locked = kwargs.get("locked", False)

    def link(self, *nodes):
        """Creates a directed link from the current node to the provided
        node(s) in the input. This function will also check that the
        linked node does not exist higher up in the DAG to the current
        node, if that is the case then a cycle has formed which breaks
        the DAG structure and could cause problems. An error will be
        thrown in this case.

        The linked node is added to a ``nodes`` dictionary that each
        node stores. This makes it easy to check which nodes are
        linked to each other.

        """
        for node in nodes:
            for subnode_id in node.flat(include_locked=True, include_links=True).keys():
                if self.identity == subnode_id:
                    raise InvalidParameter("Parameter structure must be Directed Acyclic Graph! Adding this node would create a cycle")
            self.nodes[node.name] = node
            
    def unlink(self, *nodes):
        """Undoes the linking of two nodes. Note that this could sever the
        connection of many nodes to each other if the current node was
        the only link between two branches.

        """
        for node in nodes:
            del self.nodes[node.name]

    def dump(self):
        """Simply unlinks all nodes that the current node is linked with.

        """
        self.unlink(*self.nodes.values())

    @property
    def leaf(self):
        """Returns True when the current node is a leaf node.

        """
        return len(self.nodes) == 0
    @property
    def branch(self):
        """Returns True when the current node is a branch node (not a leaf node, is linked to more nodes).

        """
        return len(self.nodes) > 0

    def __getitem__(self, key):
        """Used to get a node from the DAG relative to the current node. It
        is possible to collect nodes from deeper in the DAG by
        separating the names of the nodes along the path with a colon
        (:). For example::

          first_node["second_node:third_node"]

        returns a node that is actually linked to ``second_node``
        without needing to first get ``second_node`` then call
        ``second_node['third_node']``.

        """
        if key == self.name:
            return self
        if key in self.nodes:
            return self.nodes[key]
        if isinstance(key, str) and ":" in key:
            base, stem = key.split(":", 1)
            return self.nodes[base][stem]
        if isinstance(key, int):
            for node in self.nodes.values():
                if key == node.identity:
                    return node
        raise KeyError(f"Unrecognized key for '{self.name}': {key}")
                
    def __contains__(self, key):
        """Check if a node has a link directly to another node. A check like
        ``"second_node" in first_node`` would return true only if
        ``first_node`` was linked to ``second_node``.

        """
        return key in self.nodes

    def __eq__(self, other):
        """Equality check for nodes only returns true if they are in fact the
        same node.

        """
        return self is other

    @property
    def identity(self):
        """A read only property of the node which does not change over it's
        lifetime that uniquely identifies it relative to other
        nodes. By default this just uses the ``id(self)`` though for
        the purpose of saving/loading it may not always be this way.

        """
        try:
            return self._identity
        except AttributeError:
            return id(self)
    
    def get_state(self):
        """Returns a dictionary with state information about this node. From
        that dictionary the node can reconstruct itself, or form
        another node which is a copy of this one.

        """
        state = {
            "name": self.name,
            "identity": self.identity,
        }
        if self.locked:
            state["locked"] = self.locked
        if len(self.nodes) > 0:
            state["nodes"] = list(node.get_state() for node in self.nodes.values())
        return state

    def set_state(self, state):
        """Used to set the state of the node for the purpose of
        loading/copying. This uses the dictionary produced by
        ``get_state`` to re-create itself.

        """
        self.name = state["name"]
        self._identity = state["identity"]
        if "nodes" in state:
            for node in state["nodes"]:
                self.link(self.__class__(name = node["name"], state = node))
        self.locked = state.get("locked", False)
        
    def __iter__(self):
        return filter(lambda n: not n.locked, self.nodes.values())
    
    @property
    @abstractmethod
    def value(self):
        ...

    def flat(self, include_locked = True, include_links = False):
        """Searches the DAG from this node and collects other nodes in the
        graph. By default it will include all leaf nodes only, however
        it can be directed to only collect leaf nodes that are not
        locked, it can also be directed to collect all nodes instead
        of just leaf nodes.

        """
        flat = OrderedDict()
        if self.leaf and self.value is not None:
            if (not self.locked) or include_locked or Node.global_unlock:
                flat[self.identity] = self
        for node in self.nodes.values():
            if node.locked and not (include_locked or Node.global_unlock):
                continue
            if node.leaf and node.value is not None:
                flat[node.identity] = node
            else:
                if include_links and ((not node.locked) or include_locked or Node.global_unlock):
                    flat[node.identity] = node
                flat.update(node.flat(include_locked))
        return flat

    def __str__(self):
        return f"Node: {self.name}"
    def __repr__(self):
        return f"Node: {self.name} " + ("locked" if self.locked else "unlocked") + ("" if self.leaf else " {" + ";".join(repr(node) for node in self.nodes) + "}")



from time import time
from typing import Any, Union, Sequence, Optional

import numpy as np
import torch
from scipy.optimize import minimize
from scipy.special import gammainc

from .. import AP_config


__all__ = ["BaseOptimizer"]


class BaseOptimizer(object):
    """
    Base optimizer object that other optimizers inherit from. Ensures consistent signature for the classes.

    Parameters:
        model: an AstroPhot_Model object that will have its (unlocked) parameters optimized [AstroPhot_Model]
        initial_state: optional initialization for the parameters as a 1D tensor [tensor]
        max_iter: maximum allowed number of iterations [int]
        relative_tolerance: tolerance for counting success steps as: 0 < (Chi2^2 - Chi1^2)/Chi1^2 < tol [float]

    """

    def __init__(
        self,
        model: "AstroPhot_Model",
        initial_state: Sequence = None,
        relative_tolerance: float = 1e-3,
        fit_window: Optional["Window"] = None,
        **kwargs,
    ) -> None:
        """
        Initializes a new instance of the class.

        Args:
            model (object): An object representing the model.
            initial_state (Optional[Sequence]): The initial state of the model could be any tensor.
                           If `None`, the model's default initial state will be used.
            relative_tolerance (float): The relative tolerance for the optimization.
            fit_parameters_identity (Optiona[tuple]): a tuple of parameter identity strings which tell the LM optimizer which parameters of the model to fit.
            **kwargs (dict): Additional keyword arguments.

        Attributes:
            model (object): An object representing the model.
            verbose (int): The verbosity level.
            current_state (Tensor): The current state of the model.
            max_iter (int): The maximum number of iterations.
            iteration (int): The current iteration number.
            save_steps (Optional[str]): Save intermediate results to this path.
            relative_tolerance (float): The relative tolerance for the optimization.
            lambda_history (List[ndarray]): A list of the optimization steps.
            loss_history (List[float]): A list of the optimization losses.
            message (str): An informational message.
        """

        self.model = model
        self.verbose = kwargs.get("verbose", 0)

        if fit_window is None:
            self.fit_window = self.model.window
        else:
            self.fit_window = fit_window & self.model.window

        if initial_state is None:
            self.model.initialize()
            initial_state = self.model.parameters.vector_representation()
        else:
            initial_state = torch.as_tensor(
                initial_state, dtype=AP_config.ap_dtype, device=AP_config.ap_device
            )

        self.current_state = torch.as_tensor(
            initial_state, dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )
        if self.verbose > 1:
            AP_config.ap_logger.info(f"initial state: {self.current_state}")
        self.max_iter = kwargs.get("max_iter", 100 * len(initial_state))
        self.iteration = 0
        self.save_steps = kwargs.get("save_steps", None)

        self.relative_tolerance = relative_tolerance
        self.lambda_history = []
        self.loss_history = []
        self.message = ""

    def fit(self) -> "BaseOptimizer":
        """
        Raises:
            NotImplementedError: Error is raised if this method is not implemented in a subclass of BaseOptimizer.
        """
        raise NotImplementedError(
            "Please use a subclass of BaseOptimizer for optimization"
        )

    def step(self, current_state: torch.Tensor = None) -> None:
        """Args:
            current_state (torch.Tensor, optional): Current state of the model parameters. Defaults to None.

        Raises:
            NotImplementedError: Error is raised if this method is not implemented in a subclass of BaseOptimizer.
        """
        raise NotImplementedError(
            "Please use a subclass of BaseOptimizer for optimization"
        )

    def chi2min(self) -> float:
        """
        Returns the minimum value of chi^2 loss in the loss history.

        Returns:
                float: Minimum value of chi^2 loss.
        """
        return np.nanmin(self.loss_history)

    def res(self) -> np.ndarray:
        """Returns the value of lambda (regularization strength) at which minimum chi^2 loss was achieved.

        Returns: ndarray which is the Value of lambda at which minimum chi^2 loss was achieved.
        """
        N = np.isfinite(self.loss_history)
        if np.sum(N) == 0:
            AP_config.ap_logger.warning(
                "Getting optimizer res with no real loss history, using current state"
            )
            return self.current_state.detach().cpu().numpy()
        return np.array(self.lambda_history)[N][
            np.argmin(np.array(self.loss_history)[N])
        ]

    def res_loss(self):
        N = np.isfinite(self.loss_history)
        return np.min(np.array(self.loss_history)[N])

    @staticmethod
    def chi2contour(n_params: int, confidence: float = 0.682689492137) -> float:
        """
        Calculates the chi^2 contour for the given number of parameters.

        Args:
            n_params (int): The number of parameters.
            confidence (float, optional): The confidence interval (default is 0.682689492137).

        Returns:
            float: The calculated chi^2 contour value.

        Raises:
            RuntimeError: If unable to compute the Chi^2 contour for the given number of parameters.

        """

        def _f(x: float, nu: int) -> float:
            """Helper function for calculating chi^2 contour."""
            return (gammainc(nu / 2, x / 2) - confidence) ** 2

        for method in ["L-BFGS-B", "Powell", "Nelder-Mead"]:
            res = minimize(_f, x0=n_params, args=(n_params,), method=method, tol=1e-8)

            if res.success:
                return res.x[0]
        raise RuntimeError(f"Unable to compute Chi^2 contour for ndf: {ndf}")



import torch
import numpy as np

from .. import AP_config
from ..utils.conversions.units import deg_to_arcsec
from ..errors import SpecificationConflict, InvalidWCS

__all__ = ("WPCS", "PPCS", "WCS")

deg_to_rad = np.pi / 180
rad_to_deg = 180 / np.pi
rad_to_arcsec = rad_to_deg * 3600
arcsec_to_rad = deg_to_rad / 3600

class WPCS:
    """World to Plane Coordinate System in AstroPhot.

    AstroPhot performs its operations on a tangent plane to the
    celestial sphere, this class handles projections between the sphere and the
    tangent plane. It holds variables for the reference (RA,DEC) where
    the tangent plane contacts the sphere, and the type of projection
    being performed. Note that (RA,DEC) coordinates should always be
    in degrees while the tangent plane is in arcsecs.

    Attributes:
      reference_radec: The reference (RA,DEC) coordinates in degrees where the tangent plane contacts the sphere.
      reference_planexy: The reference tangent plane coordinates in arcsec where the tangent plane contacts the sphere.
      projection: The projection system used to convert from (RA,DEC) onto the tangent plane. Should be one of: gnomonic (default), orthographic, steriographic

    """

    # Softening length used for numerical stability and/or integration stability to avoid discontinuities (near R=0). This is in units of arcsec.
    softening = 1e-3
    
    default_reference_radec = (0,0)
    default_reference_planexy = (0,0)
    default_projection = "gnomonic"

    def __init__(self, **kwargs):
        self.projection = kwargs.get("projection", self.default_projection)
        self.reference_radec = kwargs.get("reference_radec", self.default_reference_radec)
        self.reference_planexy = kwargs.get("reference_planexy", self.default_reference_planexy)
        
        
    def world_to_plane(self, world_RA, world_DEC=None):
        """Take a coordinate on the world coordinate system, also called the
        celesial sphere, (RA, DEC in degrees) and transform it to the
        cooresponding tangent plane coordinate
        (arcsec). Transformation is done based on the chosen
        projection (default gnomonic) and reference positions. See the
        :doc:`coordinates` documentation for more details on how the
        transformation is performed.

        """
        
        if world_DEC is None:
            return torch.stack(self.world_to_plane(*world_RA))

        world_RA = torch.as_tensor(
            world_RA, dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )
        world_DEC = torch.as_tensor(
            world_DEC, dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )
        
        if self.projection == "gnomonic":
            coords = self._world_to_plane_gnomonic(
                world_RA,
                world_DEC,
            )
        elif self.projection == "orthographic":
            coords = self._world_to_plane_orthographic(
                world_RA,
                world_DEC,
            )
        elif self.projection == "steriographic":
            coords = self._world_to_plane_steriographic(
                world_RA,
                world_DEC,
            )
        return coords[0] + self.reference_planexy[0], coords[1] + self.reference_planexy[1]
    

    def plane_to_world(self, plane_x, plane_y=None):
        """Take a coordinate on the tangent plane (arcsec), and transform it
        to the cooresponding world coordinate (RA, DEC in
        degrees). Transformation is done based on the chosen
        projection (default gnomonic) and reference positions. See the
        :doc:`coordinates` documentation for more details on how the
        transformation is performed.

        """
        
        if plane_y is None:
            return torch.stack(self.plane_to_world(*plane_x))
        plane_x = torch.as_tensor(
            plane_x - self.reference_planexy[0],
            dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )
        plane_y = torch.as_tensor(
            plane_y - self.reference_planexy[1],
            dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )
        if self.projection == "gnomonic":
            return self._plane_to_world_gnomonic(
                plane_x,
                plane_y,
            )
        if self.projection == "orthographic":
            return self._plane_to_world_orthographic(
                plane_x,
                plane_y,
            )
        if self.projection == "steriographic":
            return self._plane_to_world_steriographic(
                plane_x,
                plane_y,
            )

    @property
    def projection(self):
        """
        The mathematical projection formula which described how world coordinates are mapped to the tangent plane.
        """
        return self._projection

    @projection.setter
    def projection(self, proj):
        if proj not in (
                "gnomonic",
                "orthographic",
                "steriographic",
        ):
            raise InvalidWCS(f"Unrecognized projection: {proj}. Should be one of: gnomonic, orthographic, steriographic")
        self._projection = proj

    @property
    def reference_radec(self):
        """
        RA DEC (world) coordinates where the tangent plane meets the celestial sphere. These should be in degrees.
        """
        return self._reference_radec

    @reference_radec.setter
    def reference_radec(self, radec):
        self._reference_radec = torch.as_tensor(
            radec, dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )
    @property
    def reference_planexy(self):
        """
        x y tangent plane coordinates where the tangent plane meets the celestial sphere. These should be in arcsec.
        """
        return self._reference_planexy

    @reference_planexy.setter
    def reference_planexy(self, planexy):
        self._reference_planexy = torch.as_tensor(
            planexy, dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )

    def _project_world_to_plane(self, world_RA, world_DEC):
        """
        Recurring core calculation in all the projections from world to plane.

        Args:
          world_RA: Right ascension in degrees
          world_DEC: Declination in degrees
        """
        return (
            torch.cos(world_DEC * deg_to_rad)
            * torch.sin((world_RA - self.reference_radec[0]) * deg_to_rad)
            * rad_to_arcsec,
            (
                torch.cos(self.reference_radec[1] * deg_to_rad)
                * torch.sin(world_DEC * deg_to_rad)
                - torch.sin(self.reference_radec[1] * deg_to_rad)
                * torch.cos(world_DEC * deg_to_rad)
                * torch.cos((world_RA - self.reference_radec[0]) * deg_to_rad)
            )
            * rad_to_arcsec,
        )

    def _project_plane_to_world(self, plane_x, plane_y, rho, c):
        """
        Recurring core calculation in all the projections from plane to world.

        Args:
          plane_x: tangent plane x coordinate in arcseconds.
          plane_y: tangent plane y coordinate in arcseconds.
          rho: polar radius on tangent plane.
          c: coordinate term dependent on the projection.
        """
        return (
            (
                self._reference_radec[0] * deg_to_rad
                + torch.arctan2(
                    plane_x * arcsec_to_rad * torch.sin(c),
                    rho * torch.cos(self.reference_radec[1] * deg_to_rad) * torch.cos(c)
                    - plane_y
                    * arcsec_to_rad
                    * torch.sin(self.reference_radec[1] * deg_to_rad)
                    * torch.sin(c),
                )
            )
            * rad_to_deg,
            torch.arcsin(
                torch.cos(c)
                * torch.sin(
                    self.reference_radec[1] * deg_to_rad
                )
                + plane_y
                * arcsec_to_rad
                * torch.sin(c)
                * torch.cos(self.reference_radec[1] * deg_to_rad)
                / rho
            )
            * rad_to_deg,
        )

    def _world_to_plane_gnomonic(self, world_RA, world_DEC):
        """Gnomonic projection: (RA,DEC) to tangent plane.

        Performs Gnomonic projection of (RA,DEC) coordinates onto a
        tangent plane. The tangent plane makes contact at the location
        of the `reference_radec` variable. In a gnomonic projection,
        great circles are mapped to straight lines. The gnomonic
        projection represents the image formed by a spherical lens,
        and is sometimes known as the rectilinear projection.

        Args:
          world_RA: Right ascension in degrees
          world_DEC: Declination in degrees

        See: https://mathworld.wolfram.com/GnomonicProjection.html

        """
        C = torch.sin(self.reference_radec[1] * deg_to_rad) * torch.sin(
            world_DEC * deg_to_rad
        ) + torch.cos(self.reference_radec[1] * deg_to_rad) * torch.cos(
            world_DEC * deg_to_rad
        ) * torch.cos(
            (world_RA - self.reference_radec[0]) * deg_to_rad
        )
        x, y = self._project_world_to_plane(world_RA, world_DEC)
        return x / C, y / C

    def _plane_to_world_gnomonic(self, plane_x, plane_y):
        """Inverse Gnomonic projection: tangent plane to (RA,DEC).

        Performs the inverse Gnomonic projection of tangent plane
        coordinates into (RA,DEC) coordinates. The tangent plane makes
        contact at the location of the `reference_radec` variable. In
        a gnomonic projection, great circles are mapped to straight
        lines. The gnomonic projection represents the image formed by
        a spherical lens, and is sometimes known as the rectilinear
        projection.

        Args:
          plane_x: tangent plane x coordinate in arcseconds.
          plane_y: tangent plane y coordinate in arcseconds.

        See: https://mathworld.wolfram.com/GnomonicProjection.html

        """
        rho = (torch.sqrt(plane_x ** 2 + plane_y ** 2) + self.softening) * arcsec_to_rad
        c = torch.arctan(rho)

        ra, dec = self._project_plane_to_world(plane_x, plane_y, rho, c)
        return ra, dec

    def _world_to_plane_steriographic(self, world_RA, world_DEC):
        """Steriographic projection: (RA,DEC) to tangent plane

        Performs Steriographic projection of (RA,DEC) coordinates onto
        a tangent plane. The tangent plane makes contact at the
        location of the `reference_radec` variable. The steriographic
        projection preserves circles and angle measures.

        Args:
          world_RA: Right ascension in degrees
          world_DEC: Declination in degrees

        See: https://mathworld.wolfram.com/StereographicProjection.html

        """
        C = (
            1
            + torch.sin(world_DEC * deg_to_rad)
            * torch.sin(self._reference_radec[1] * deg_to_rad)
            + torch.cos(world_DEC * deg_to_rad)
            * torch.cos(self._reference_radec[1] * deg_to_rad)
            * torch.cos((world_RA - self._reference_radec[0]) * deg_to_rad)
        ) / 2
        x, y = self._project_world_to_plane(world_RA, world_DEC)
        return x / C, y / C

    def _plane_to_world_steriographic(self, plane_x, plane_y):
        """Inverse Steriographic projection: tangent plane to (RA,DEC).

        Performs the inverse Steriographic projection of tangent plane
        coordinates into (RA,DEC) coordinates. The tangent plane makes
        contact at the location of the `reference_radec` variable. The
        steriographic projection preserves circles and angle measures.

        Args:
          plane_x: tangent plane x coordinate in arcseconds. The origin of the tangent plane is the contact point with the sphere, represented by `reference_radec`.
          plane_y: tangent plane y coordinate in arcseconds. The origin of the tangent plane is the contact point with the sphere, represented by `reference_radec`.

        See: https://mathworld.wolfram.com/StereographicProjection.html

        """
        rho = (torch.sqrt(plane_x ** 2 + plane_y ** 2) + self.softening) * arcsec_to_rad
        c = 2 * torch.arctan(rho / 2)
        ra, dec = self._project_plane_to_world(plane_x, plane_y, rho, c)
        return ra, dec

    def _world_to_plane_orthographic(self, world_RA, world_DEC):
        """Orthographic projection: (RA,DEC) to tangent plane

        Performs Orthographic projection of (RA,DEC) coordinates onto
        a tangent plane. The tangent plane makes contact at the
        location of the `reference_radec` variable. The point of
        perspective for the orthographic projection is at infinite
        distance. This projection is perhaps better suited to
        represent the view of an exoplanet, however it is included
        here for completeness.

        Args:
          world_RA: Right ascension in degrees
          world_DEC: Declination in degrees

        See: https://mathworld.wolfram.com/OrthographicProjection.html

        """
        x, y = self._project_world_to_plane(world_RA, world_DEC)
        return x, y

    def _plane_to_world_orthographic(self, plane_x, plane_y):
        """Inverse Orthographic projection: tangent plane to (RA,DEC).

        Performs the inverse Orthographic projection of tangent plane
        coordinates into (RA,DEC) coordinates. The tangent plane makes
        contact at the location of the `reference_radec` variable. The
        point of perspective for the orthographic projection is at
        infinite distance. This projection is perhaps better suited to
        represent the view of an exoplanet, however it is included
        here for completeness.

        Args:
          plane_x: tangent plane x coordinate in arcseconds. The origin of the tangent plane is the contact point with the sphere, represented by `reference_radec`.
          plane_y: tangent plane y coordinate in arcseconds. The origin of the tangent plane is the contact point with the sphere, represented by `reference_radec`.

        See: https://mathworld.wolfram.com/OrthographicProjection.html

        """
        rho = (torch.sqrt(plane_x ** 2 + plane_y ** 2) + self.softening) * arcsec_to_rad
        c = torch.arcsin(rho)

        ra, dec = self._project_plane_to_world(plane_x, plane_y, rho, c)
        return ra, dec

    def get_state(self):
        """Returns a dictionary with the information needed to recreate the
        WPCS object.

        """
        return {
            "projection": self.projection,
            "reference_radec": self.reference_radec.detach().cpu().tolist(),
            "reference_planexy": self.reference_planexy.detach().cpu().tolist(),
        }
    def set_state(self, state):
        """Takes a state dictionary and re-creates the state of the WPCS
        object.

        """
        self.projection = state.get("projection", self.default_projection)
        self.reference_radec = state.get("reference_radec", self.default_reference_radec)
        self.reference_planexy = state.get("reference_planexy", self.default_reference_planexy)

    def get_fits_state(self):
        """
        Similar to get_state, except specifically tailored to be stored in a FITS format.
        """
        return {
            "PROJ": self.projection,
            "REFRADEC": str(self.reference_radec.detach().cpu().tolist()),
            "REFPLNXY": str(self.reference_planexy.detach().cpu().tolist()),
        }
    
    def set_fits_state(self, state):
        """
        Reads and applies the state from the get_fits_state function.
        """
        self.projection = state["PROJ"]
        self.reference_radec = eval(state["REFRADEC"])
        self.reference_planexy = eval(state["REFPLNXY"])
        
    def copy(self, **kwargs):
        """Create a copy of the WPCS object with the same projection
        paramaters.

        """
        copy_kwargs = {
            "projection": self.projection,
            "reference_radec": self.reference_radec,
            "reference_planexy": self.reference_planexy,
        }
        copy_kwargs.update(kwargs)
        return self.__class__(
            **copy_kwargs,
        )
    
    def to(self, dtype=None, device=None):
        """
        Convert all stored tensors to a new device and data type
        """
        if dtype is None:
            dtype = AP_config.ap_dtype
        if device is None:
            device = AP_config.ap_device
        self._reference_radec = self._reference_radec.to(dtype=dtype, device=device)
        self._reference_planexy = self._reference_planexy.to(dtype=dtype, device=device)

    def __str__(self):
        return f"WPCS reference_radec: {self.reference_radec.detach().cpu().tolist()}, reference_planexy: {self.reference_planexy.detach().cpu().tolist()}"
    def __repr__(self):
        return f"WPCS reference_radec: {self.reference_radec.detach().cpu().tolist()}, reference_planexy: {self.reference_planexy.detach().cpu().tolist()}, projection: {self.projection}"

class PPCS:
    """
    plane to pixel coordinate system


    Args:
      pixelscale : float or None, optional
          The physical scale of the pixels in the image, this is
          represented as a matrix which projects pixel units into sky
          units: :math:`pixelscale @ pixel_vec = sky_vec`. The pixel
          scale matrix can be thought of in four components:
          :math:`\vec{s} @ F @ R @ S` where :math:`\vec{s}` is the side
          length of the pixels, :math:`F` is a diagonal matrix of {1,-1}
          which flips the axes orientation, :math:`R` is a rotation
          matrix, and :math:`S` is a shear matrix which turns
          rectangular pixels into parallelograms. Default is None.      
      reference_imageij : Sequence or None, optional
          The pixel coordinate at which the image is fixed to the
          tangent plane. By default this is (-0.5, -0.5) or the bottom
          corner of the [0,0] indexed pixel.
      reference_imagexy : Sequence or None, optional
          The tangent plane coordinate at which the image is fixed,
          corresponding to the reference_imageij coordinate. These two
          reference points ar pinned together, any rotations would occur
          about this point. By default this is (0., 0.).
    
    """
    
    default_reference_imageij = (-0.5,-0.5)
    default_reference_imagexy = (0,0)
    default_pixelscale = 1

    def __init__(self, *, wcs=None, pixelscale=None, **kwargs):

        self.reference_imageij = kwargs.get("reference_imageij", self.default_reference_imageij)        
        self.reference_imagexy = kwargs.get("reference_imagexy", self.default_reference_imagexy)

        # Collect the pixelscale of the pixel grid
        if wcs is not None and pixelscale is None:
            self.pixelscale = deg_to_arcsec * wcs.pixel_scale_matrix
        elif pixelscale is not None:
            if wcs is not None and isinstance(pixelscale, float):
                AP_config.ap_logger.warning(
                    "Overriding WCS pixelscale with manual input! To remove this message, either let WCS define pixelscale, or input full pixelscale matrix"
                )
            self.pixelscale = pixelscale
        else:
            AP_config.ap_logger.warning(
                "Assuming pixelscale of 1! To remove this message please provide the pixelscale explicitly"
            )
            self.pixelscale = self.default_pixelscale
        
    @property
    def pixelscale(self):
        """Matrix defining the shape of pixels in the tangent plane, these
        can be any parallelogram defined by the matrix.

        """
        return self._pixelscale

    @pixelscale.setter
    def pixelscale(self, pix):
        if pix is None:
            self._pixelscale = None
            return

        self._pixelscale = (
            torch.as_tensor(pix, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
            .clone()
            .detach()
        )
        if self._pixelscale.numel() == 1:
            self._pixelscale = torch.tensor(
                [[self._pixelscale.item(), 0.0], [0.0, self._pixelscale.item()]],
                dtype=AP_config.ap_dtype,
                device=AP_config.ap_device,
            )
        self._pixel_area = torch.linalg.det(self.pixelscale).abs()
        self._pixel_length = self._pixel_area.sqrt()
        self._pixelscale_inv = torch.linalg.inv(self.pixelscale)

    @property
    def pixel_area(self):
        """The area inside a pixel in arcsec^2

        """
        return self._pixel_area

    @property
    def pixel_length(self):
        """The approximate length of a pixel, which is just
        sqrt(pixel_area). For square pixels this is the actual pixel
        length, for rectangular pixels it is a kind of average.

        The pixel_length is typically not used for exact calculations
        and instead sets a size scale within an image.

        """
        return self._pixel_length

    @property
    def reference_imageij(self):
        """pixel coordinates where the pixel grid is fixed to the tangent
        plane. These should be in pixel units where (0,0) is the
        center of the [0,0] indexed pixel. However, it is still in xy
        format, meaning that the first index gives translations in the
        x-axis (horizontal-axis) of the image.

        """
        return self._reference_imageij

    @reference_imageij.setter
    def reference_imageij(self, imageij):
        self._reference_imageij = torch.as_tensor(
            imageij, dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )
    @property
    def reference_imagexy(self):
        """plane coordinates where the image grid is fixed to the tangent
        plane. These should be in arcsec.

        """
        return self._reference_imagexy

    @reference_imagexy.setter
    def reference_imagexy(self, imagexy):
        self._reference_imagexy = torch.as_tensor(
            imagexy, dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )

    def pixel_to_plane(self, pixel_i, pixel_j=None):
        """Take in a coordinate on the regular pixel grid, where 0,0 is the
        center of the [0,0] indexed pixel. This coordinate is
        transformed into the tangent plane coordinate system (arcsec)
        based on the pixel scale and reference positions. If the pixel
        scale matrix is :math:`P`, the reference pixel is
        :math:`\vec{r}_{pix}`, the reference tangent plane point is
        :math:`\vec{r}_{tan}`, and the coordinate to transform is
        :math:`\vec{c}_{pix}` then the coordinate in the tangent plane
        is:

        .. math::
            \vec{c}_{tan} = [P(\vec{c}_{pix} - \vec{r}_{pix})] + \vec{r}_{tan}

        """
        if pixel_j is None:
            return torch.stack(self.pixel_to_plane(*pixel_i))
        coords = torch.mm(self.pixelscale, torch.stack((pixel_i.reshape(-1), pixel_j.reshape(-1))) - self.reference_imageij.view(2,1)) + self.reference_imagexy.view(2,1)
        return coords[0].reshape(pixel_i.shape), coords[1].reshape(pixel_j.shape)

    def plane_to_pixel(self, plane_x, plane_y=None):
        """Take a coordinate on the tangent plane (arcsec) and transform it to
        the cooresponding pixel grid coordinate (pixel units where
        (0,0) is the [0,0] indexed pixel). Transformation is done
        based on the pixel scale and reference positions. If the pixel
        scale matrix is :math:`P`, the reference pixel is
        :math:`\vec{r}_{pix}`, the reference tangent plane point is
        :math:`\vec{r}_{tan}`, and the coordinate to transform is
        :math:`\vec{c}_{tan}` then the coordinate in the pixel grid
        is:

        .. math::
            \vec{c}_{pix} = [P^{-1}(\vec{c}_{tan} - \vec{r}_{tan})] + \vec{r}_{pix}

        """
        if plane_y is None:
            return torch.stack(self.plane_to_pixel(*plane_x))
        coords = torch.mm(self._pixelscale_inv, torch.stack((plane_x.reshape(-1), plane_y.reshape(-1))) - self.reference_imagexy.view(2,1)) + self.reference_imageij.view(2,1)
        return coords[0].reshape(plane_x.shape), coords[1].reshape(plane_y.shape)

    def pixel_to_plane_delta(self, pixel_delta_i, pixel_delta_j=None):
        """Take a translation in pixel space and determine the cooresponding
        translation in the tangent plane (arcsec). Essentially this performs
        the pixel scale matrix multiplication without any reference
        coordinates applied.

        """
        if pixel_delta_j is None:
            return torch.stack(self.pixel_to_plane_delta(*pixel_delta_i))
        coords = torch.mm(self.pixelscale, torch.stack((pixel_delta_i.reshape(-1), pixel_delta_j.reshape(-1))))
        return coords[0].reshape(pixel_delta_i.shape), coords[1].reshape(pixel_delta_j.shape)

    def plane_to_pixel_delta(self, plane_delta_x, plane_delta_y=None):
        """Take a translation in tangent plane space (arcsec) and determine
        the cooresponding translation in pixel space. Essentially this
        performs the pixel scale matrix multiplication without any
        reference coordinates applied.

        """
        if plane_delta_y is None:
            return torch.stack(self.plane_to_pixel_delta(*plane_delta_x))
        coords = torch.mm(self._pixelscale_inv, torch.stack((plane_delta_x.reshape(-1), plane_delta_y.reshape(-1))))
        return coords[0].reshape(plane_delta_x.shape), coords[1].reshape(plane_delta_y.shape)
        
    def copy(self, **kwargs):
        """Create a copy of the PPCS object with the same projection
        paramaters.

        """
        copy_kwargs = {
            "pixelscale": self.pixelscale,
            "reference_imageij": self.reference_imageij,
            "reference_imagexy": self.reference_imagexy,
        }
        copy_kwargs.update(kwargs)
        return self.__class__(
            **copy_kwargs,
        )
    
    def get_state(self):
        return {
            "pixelscale": self.pixelscale.detach().cpu().tolist(),
            "reference_imageij": self.reference_imageij.detach().cpu().tolist(),
            "reference_imagexy": self.reference_imagexy.detach().cpu().tolist(),
        }
    def set_state(self, state):
        self.pixelscale = state.get("pixelscale", self.default_pixelscale)
        self.reference_imageij = state.get("reference_imageij", self.default_reference_imageij)
        self.reference_imagexy = state.get("reference_imagexy", self.default_reference_imagexy)
    
    def get_fits_state(self):
        """
        Similar to get_state, except specifically tailored to be stored in a FITS format.
        """
        return {
            "PXLSCALE": str(self.pixelscale.detach().cpu().tolist()),
            "REFIMGIJ": str(self.reference_imageij.detach().cpu().tolist()),
            "REFIMGXY": str(self.reference_imagexy.detach().cpu().tolist()),
        }
    
    def set_fits_state(self, state):
        """
        Reads and applies the state from the get_fits_state function.
        """
        self.pixelscale = eval(state["PXLSCALE"])
        self.reference_imageij = eval(state["REFIMGIJ"])
        self.reference_imagexy = eval(state["REFIMGXY"])
        
    def to(self, dtype=None, device=None):
        """
        Convert all stored tensors to a new device and data type
        """
        if dtype is None:
            dtype = AP_config.ap_dtype
        if device is None:
            device = AP_config.ap_device
        self._pixelscale = self._pixelscale.to(dtype=dtype, device=device)
        self._reference_imageij = self._reference_imageij.to(dtype=dtype, device=device)
        self._reference_imagexy = self._reference_imagexy.to(dtype=dtype, device=device)

    def __str__(self):
        return f"PPCS reference_imageij: {self.reference_imageij.detach().cpu().tolist()}, reference_imagexy: {self.reference_imagexy.detach().cpu().tolist()}"
    def __repr__(self):
        return f"PPCS reference_imageij: {self.reference_imageij.detach().cpu().tolist()}, reference_imagexy: {self.reference_imagexy.detach().cpu().tolist()}, pixelscale: {self.pixelscale.detach().cpu().tolist()}"

        
class WCS(WPCS, PPCS):
    """
    Full world coordinate system defines mappings from world to tangent plane to pixel grid and all other variations.
    """

    def __init__(self, *args, wcs=None, **kwargs):
        if kwargs.get("state", None) is not None:
            self.set_state(kwargs["state"])
            return
        
        if wcs is not None:
            if wcs.wcs.ctype[0] != "RA---TAN":
                AP_config.ap_logger.warning("Astropy WCS not tangent plane coordinate system! May not be compatible with AstroPhot.")
            if wcs.wcs.ctype[1] != "DEC--TAN":
                AP_config.ap_logger.warning("Astropy WCS not tangent plane coordinate system! May not be compatible with AstroPhot.")
                
        if wcs is not None:
            kwargs["reference_radec"] = kwargs.get("reference_radec", wcs.wcs.crval)
            kwargs["reference_imageij"] = wcs.wcs.crpix
            WPCS.__init__(self, *args, wcs=wcs, **kwargs)
            sky_coord = wcs.pixel_to_world(*wcs.wcs.crpix)
            kwargs["reference_imagexy"] = self.world_to_plane(torch.tensor((sky_coord.ra.deg, sky_coord.dec.deg), dtype=AP_config.ap_dtype, device=AP_config.ap_device))
        else:
            WPCS.__init__(self, *args, **kwargs)
            
        PPCS.__init__(self, *args, wcs=wcs, **kwargs)

    def world_to_pixel(self, world_RA, world_DEC=None):
        """A wrapper which applies :meth:`world_to_plane` then
        :meth:`plane_to_pixel`, see those methods for further
        information.

        """
        if world_DEC is None:
            return torch.stack(self.world_to_pixel(*world_RA))
        return self.plane_to_pixel(*self.world_to_plane(world_RA, world_DEC))
    
    def pixel_to_world(self, pixel_i, pixel_j=None):
        """A wrapper which applies :meth:`pixel_to_plane` then
        :meth:`plane_to_world`, see those methods for further
        information.

        """
        if pixel_j is None:
            return torch.stack(self.pixel_to_world(*pixel_i))
        return self.plane_to_world(*self.pixel_to_plane(pixel_i, pixel_j))

    def copy(self, **kwargs):
        copy_kwargs = {
            "pixelscale": self.pixelscale,
            "reference_imageij": self.reference_imageij,
            "reference_imagexy": self.reference_imagexy,
            "projection": self.projection,
            "reference_radec": self.reference_radec,
            "reference_planexy": self.reference_planexy,
        }
        copy_kwargs.update(kwargs)
        return self.__class__(
            **copy_kwargs,     
        )

    def to(self, dtype=None, device=None):
        WPCS.to(self, dtype, device)
        PPCS.to(self, dtype, device)

    def get_state(self):
        state = WPCS.get_state(self)
        state.update(PPCS.get_state(self))
        return state

    def set_state(self, state):
        WPCS.set_state(self, state)
        PPCS.set_state(self, state)
        
    def get_fits_state(self):
        """
        Similar to get_state, except specifically tailored to be stored in a FITS format.
        """
        state = WPCS.get_fits_state(self)
        state.update(PPCS.get_fits_state(self))
        return state
    
    def set_fits_state(self, state):
        """
        Reads and applies the state from the get_fits_state function.
        """
        WPCS.set_fits_state(self, state)
        PPCS.set_fits_state(self, state)
        
    def __str__(self):
        return f"WCS:\n{WPCS.__str__(self)}\n{PPCS.__str__(self)}"
    def __repr__(self):
        return f"WCS:\n{WPCS.__repr__(self)}\n{PPCS.__repr__(self)}"



from .image_object import *
from .image_header import *
from .target_image import *
from .jacobian_image import *
from .psf_image import *
from .model_image import *
from .window_object import *
from .wcs import *



import torch
import numpy as np

from .psf_model_object import PSF_Model
from ..image import PSF_Image
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.interpolate import interp2d
from ._shared_methods import select_target
from ..param import Param_Unlock, Param_SoftLimits
from .. import AP_config

__all__ = ["Eigen_PSF"]

class Eigen_PSF(PSF_Model):
    """point source model which uses multiple images as a basis for the
    PSF as its representation for point sources. Using bilinear
    interpolation it will shift the PSF within a pixel to accurately
    represent the center location of a point source. There is no
    functional form for this object type as any image can be
    supplied. Note that as an argument to the model at construction
    one can provide "psf" as an AstroPhot PSF_Image object. Since only
    bilinear interpolation is performed, it is recommended to provide
    the PSF at a higher resolution than the image if it is near the
    nyquist sampling limit. Bilinear interpolation is very fast and
    accurate for smooth models, so this way it is possible to do the
    expensive interpolation before optimization and save time. Note
    that if you do this you must provide the PSF as a PSF_Image object
    with the correct pixelscale (essentially just divide the
    pixelscale by the upsampling factor you used).

    Args:
      eigen_basis (tensor): This is the basis set of images used to form the eigen point source, it should be a tensor with shape (N x W x H) where N is the number of eigen images, and W/H are the dimensions of the image.
      eigen_pixelscale (float): This is the pixelscale associated with the eigen basis images.
    
    Parameters:
        flux: the total flux of the point source model, represented as the log of the total flux.
        weights: the relative amplitude of the Eigen basis modes.

    """

    model_type = f"eigen {PSF_Model.model_type}"
    parameter_specs = {
        "flux": {"units": "log10(flux/arcsec^2)", "value": 0., "locked": True},
        "weights": {"units": "unitless"},
    }
    _parameter_order = PSF_Model._parameter_order + ("flux","weights")
    useable = True
    model_integrated = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "eigen_basis" not in kwargs:
            AP_config.ap_logger.warning("Eigen basis not supplied! Assuming psf as single basis element. Please provide Eigen basis or just use an empirical PSF image.")
            self.eigen_basis = torch.clone(self.target.data).unsqueeze(0)
            self.parameters["weights"].locked = True
        else:
            self.eigen_basis = torch.as_tensor(
                kwargs["eigen_basis"],
                dtype = AP_config.ap_dtype,
                device = AP_config.ap_device
            )
        if kwargs.get("normalize_eigen_basis", True):
            self.eigen_basis = self.eigen_basis / torch.sum(self.eigen_basis, axis = (1,2)).unsqueeze(1).unsqueeze(2)
        self.eigen_pixelscale = torch.as_tensor(
            kwargs.get("eigen_pixelscale", 1. if self.target is None else self.target.pixelscale),
            dtype = AP_config.ap_dtype,
            device = AP_config.ap_device
        )
        
    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)
        target_area = target[self.window]
        with Param_Unlock(parameters["flux"]), Param_SoftLimits(parameters["flux"]):
            if parameters["flux"].value is None:
                parameters["flux"].value = torch.log10(torch.abs(torch.sum(target_area.data)) / target.pixel_area)
            if parameters["flux"].uncertainty is None:
                parameters["flux"].uncertainty = torch.abs(parameters["flux"].value) * self.default_uncertainty
        with Param_Unlock(parameters["weights"]), Param_SoftLimits(parameters["weights"]):
            if parameters["weights"].value is None:
                W = np.zeros(len(self.eigen_basis))
                W[0] = 1.
                parameters["weights"].value = W
            if parameters["weights"].uncertainty is None:
                parameters["weights"].uncertainty = torch.ones_like(parameters["weights"].value) * self.default_uncertainty

    @default_internal
    def evaluate_model(self, X=None, Y=None, image=None, parameters=None, **kwargs):
        if X is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]

        psf_model = PSF_Image(
            data = torch.clamp(torch.sum(self.eigen_basis.detach() * (parameters["weights"].value / torch.linalg.norm(parameters["weights"].value)).unsqueeze(1).unsqueeze(2), axis = 0), min = 0.),
            pixelscale = self.eigen_pixelscale.detach(),
        )
        
        # Convert coordinates into pixel locations in the psf image
        pX, pY = psf_model.plane_to_pixel(X, Y)

        # Select only the pixels where the PSF image is defined
        select = torch.logical_and(
            torch.logical_and(pX > -0.5, pX < psf_model.data.shape[1]-0.5),
            torch.logical_and(pY > -0.5, pY < psf_model.data.shape[0]-0.5),
        )

        # Zero everywhere outside the psf
        result = torch.zeros_like(X)

        # Use bilinear interpolation of the PSF at the requested coordinates
        result[select] = interp2d(psf_model.data, pX[select], pY[select])

        # Ensure positive values
        result = torch.clamp(result, min = 0.)
        
        return result * (image.pixel_area * 10 ** parameters["flux"].value)
